# Copyright (c) nexB, Inc. and others.
# Copyright (c) Center for Information Technology, http://coi.gov.pl
# Copyright (c) Chad Fowler, Rich Kilmer, Jim Weirich and others.
# Copyright (c) Engine Yard and Andre Arko, Facebook, Inc. and its affiliates.
#
# SPDX-License-Identifier: Apache-2.0 AND MIT
# This has been significantly modified from the original
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

# notes: This has been substantially modified and enhanced from the original
# puppeteer code to extract the Ruby version handling code.
# This code is in turn originally based and updated from the Rubygems code
# Originally from https://github.com/rubygems/rubygems and
# https://github.com/coi-gov-pl/puppeter

import operator
import re
from collections import namedtuple
from itertools import dropwhile


class InvalidRequirementError(AttributeError):
    pass


class InvalidVersionError(ValueError):
    pass


class GemVersion:
    """
    The Rubygems version.rb has this documentation
    The Version class processes string versions into comparable
    values. A version string should normally be a series of numbers
    separated by periods. Each part (digits separated by periods) is
    considered its own number, and these are used for sorting. So for
    instance, 3.10 sorts higher than 3.2 because ten is greater than
    two.

    If any part contains letters (currently only a-z are supported) then
    that version is considered prerelease. Versions with a prerelease
    part in the Nth part sort less than versions with N-1
    parts. Prerelease parts are sorted alphabetically using the normal
    Ruby string sorting rules. If a prerelease part contains both
    letters and numbers, it will be broken into multiple parts to
    provide expected sort behavior (1.0.a10 becomes 1.0.a.10, and is
    greater than 1.0.a9).

    Prereleases sort between real releases (newest to oldest):

    1. 1.0
    2. 1.0.b1
    3. 1.0.a.2
    4. 0.9

    If you want to specify a version restriction that includes both prereleases
    and regular releases of the 1.x series this is the best way:

      s.add_dependency 'example', '>= 1.0.0.a', '< 2.0.0'

    == How Software Changes

    Users expect to be able to specify a version constraint that gives them
    some reasonable expectation that new versions of a library will work with
    their software if the version constraint is true, and not work with their
    software if the version constraint is false.  In other words, the perfect
    system will accept all compatible versions of the library and reject all
    incompatible versions.

    Libraries change in 3 ways (well, more than 3, but stay focused here!).

    1. The change may be an implementation detail only and have no effect on
       the client software.
    2. The change may add new features, but do so in a way that client software
       written to an earlier version is still compatible.
    3. The change may change the public interface of the library in such a way
       that old software is no longer compatible.

    Some examples are appropriate at this point.  Suppose I have a Stack class
    that supports a <tt>push</tt> and a <tt>pop</tt> method.

    === Examples of Category 1 changes:

    * Switch from an array based implementation to a linked-list based
      implementation.
    * Provide an automatic (and transparent) backing store for large stacks.

    === Examples of Category 2 changes might be:

    * Add a <tt>depth</tt> method to return the current depth of the stack.
    * Add a <tt>top</tt> method that returns the current top of stack (without
      changing the stack).
    * Change <tt>push</tt> so that it returns the item pushed (previously it
      had no usable return value).

    === Examples of Category 3 changes might be:

    * Changes <tt>pop</tt> so that it no longer returns a value (you must use
      <tt>top</tt> to get the top of the stack).
    * Rename the methods to <tt>push_item</tt> and <tt>pop_item</tt>.

    == RubyGems Rational Versioning

    * Versions shall be represented by three non-negative integers, separated
      by periods (e.g. 3.1.4).  The first integers is the "major" version
      number, the second integer is the "minor" version number, and the third
      integer is the "build" number.

    * A category 1 change (implementation detail) will increment the build
      number.

    * A category 2 change (backwards compatible) will increment the minor
      version number and reset the build number.

    * A category 3 change (incompatible) will increment the major build number
      and reset the minor and build numbers.

    * Any "public" release of a gem should have a different version.  Normally
      that means incrementing the build number.  This means a developer can
      generate builds all day long, but as soon as they make a public release,
      the version must be updated.

    === Examples

    Let's work through a project lifecycle using our Stack example from above.

    Version 0.0.1:: The initial Stack class is release.
    Version 0.0.2:: Switched to a linked=list implementation because it is
                    cooler.
    Version 0.1.0:: Added a <tt>depth</tt> method.
    Version 1.0.0:: Added <tt>top</tt> and made <tt>pop</tt> return nil
                    (<tt>pop</tt> used to return the  old top item).
    Version 1.1.0:: <tt>push</tt> now returns the value pushed (it used it
                    return nil).
    Version 1.1.1:: Fixed a bug in the linked list implementation.
    Version 1.1.2:: Fixed a bug introduced in the last fix.

    Client A needs a stack with basic push/pop capability.  They write to the
    original interface (no <tt>top</tt>), so their version constraint looks like:

      gem 'stack', '>= 0.0'

    Essentially, any version is OK with Client A.  An incompatible change to
    the library will cause them grief, but they are willing to take the chance
    (we call Client A optimistic).

    Client B is just like Client A except for two things: (1) They use the
    <tt>depth</tt> method and (2) they are worried about future
    incompatibilities, so they write their version constraint like this:

      gem 'stack', '~> 0.1'

    The <tt>depth</tt> method was introduced in version 0.1.0, so that version
    or anything later is fine, as long as the version stays below version 1.0
    where incompatibilities are introduced.  We call Client B pessimistic
    because they are worried about incompatible future changes (it is OK to be
    pessimistic!).

    == Preventing Version Catastrophe:

    From: http://blog.zenspider.com/2008/10/rubygems-howto-preventing-cata.html

    Let's say you're depending on the fnord gem version 2.y.z. If you
    specify your dependency as ">= 2.0.0" then, you're good, right? What
    happens if fnord 3.0 comes out and it isn't backwards compatible
    with 2.y.z? Your stuff will break as a result of using ">=". The
    better route is to specify your dependency with an "approximate" version
    specifier ("~>"). They're a tad confusing, so here is how the dependency
    specifiers work:

      Specification From  ... To (exclusive)
      ">= 3.0"      3.0   ... &infin;
      "~> 3.0"      3.0   ... 4.0
      "~> 3.0.0"    3.0.0 ... 3.1
      "~> 3.5"      3.5   ... 4.0
      "~> 3.5.0"    3.5.0 ... 3.6
      "~> 3"        3.0   ... 4.0

    For the last example, single-digit versions are automatically extended with
    a zero to give a sensible result.
    """

    VERSION_PATTERN = r"[0-9]+(?:\.[0-9a-zA-Z]+)*(-[0-9A-Za-z-]+(\.[0-9A-Za-z-]+)*)?"
    is_correct = re.compile(rf"^\s*({VERSION_PATTERN})?\s*$").match

    def __init__(self, version):
        """
        Construct a Version from the ``version`` string.  A version string is a
        series of digits or ASCII letters separated by dots and may contain dash
        "-".
        """
        if isinstance(version, (int, GemVersion)):
            version = str(version)

        if not isinstance(version, str):
            raise InvalidVersionError(version)

        if not self.is_correct(version):
            raise InvalidVersionError(version)

        # If version is an empty string convert it to 0
        version = str(version).strip()

        self.original = version

        if not version:
            version = "0"

        self.version = version.replace("-", ".pre.")
        self._segments = ()
        self._canonical_segments = ()
        self._bump = None
        self._release = None

    def __str__(self):
        return self.original

    to_string = __str__

    def __repr__(self):
        return f"GemVersion({self.original!r})"

    def equal_strictly(self, other):
        return self.version == other.version

    def __hash__(self):
        return hash(self.canonical_segments)

    def __eq__(self, other):
        return self.canonical_segments == other.canonical_segments

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __le__(self, other):
        return self.__cmp__(other) <= 0

    def __gt__(self, other):
        return self.__cmp__(other) > 0

    def __ge__(self, other):
        return self.__cmp__(other) >= 0

    def bump(self):
        """
        Return a new version object where the next to the last revision number
        is one greater (e.g., 5.3.1 => 5.4) i.e., incrementing this GemVersion
        last numeric segment.

        For example::
        >>> assert GemVersion("5.3.1").bump() == GemVersion("5.4"), repr(GemVersion("5.3.1").bump())
        >>> assert GemVersion("5.3.1.4-2").bump() == GemVersion("5.3.2"), GemVersion("5.3.1.4-2").bump()
        """
        if not self._bump:
            segments = []
            for seg in self.segments:
                if isinstance(seg, str):
                    break
                else:
                    segments.append(seg)

            if len(segments) > 1:
                segments.pop()

            segments[-1] += 1
            segments = [str(r) for r in segments]
            self._bump = GemVersion(".".join(segments))

        return self._bump

    def release(self):
        """
        Return a new GemVersion which is the release for this version (e.g.,
        1.2.0.a -> 1.2.0). Non-prerelease versions return themselves. A release
        is composed only of numeric segments.
        """
        if not self._release:
            if self.prerelease():
                segments = self.segments
                while any(isinstance(s, str) for s in segments):
                    segments.pop()
                segments = (str(s) for s in segments)
                self._release = GemVersion(".".join(segments))
            else:
                self._release = self

        return self._release

    def prerelease(self):
        """
        Return True if this is considered as a prerelease version.
        A version is considered a prerelease if it contains a letter.
        """
        return any(not str(s).isdigit() for s in self.segments)

    @property
    def segments(self):
        """
        Return a new sequence of segments for this version where segments are
        ints or strings parsed from the original version string.
        """
        if not self._segments:
            self._segments = self.get_segments()
        return list(self._segments)

    def get_segments(self):
        """
        Return a sequence of segments for this version where segments are ints
        or strings parsed from the original version string.
        """
        find_segments = re.compile(r"[0-9]+|[a-z]+", re.IGNORECASE).findall
        segments = []
        for seg in find_segments(self.version):
            if seg.isdigit():
                seg = int(seg)
            segments.append(seg)
        return tuple(segments)

    @property
    def canonical_segments(self):
        if not self._canonical_segments:
            self._canonical_segments = self.get_canonical_segments()
        return list(self._canonical_segments)

    def get_canonical_segments(self):
        """
        Return a new sequence of "canonical segments" for this version using
        the Rubygems way.
        """
        canonical_segments = []
        for segments in self.split_segments():
            segs = list(dropwhile(lambda s: s == 0, reversed(segments)))
            segs = reversed(segs)
            canonical_segments.extend(segs)
        return tuple(canonical_segments)

    def split_segments(self):
        """
         Return a two-tuple of segments:
        - the first is a list of numeric-only segments starting from the left
        - the second is a list of alpha or numericsegments starting with the
          first alpha segment from the left.
        """
        numeric_segments = []
        string_segments = []
        for seg in self.segments:
            is_numeric = isinstance(seg, int)  # or (isinstance(seg, str) and seg.isdigit())
            if is_numeric:
                if string_segments:
                    string_segments.append(seg)
                else:
                    numeric_segments.append(seg)
            else:
                string_segments.append(seg)
        return numeric_segments, string_segments

    def __cmp__(self, other, trace=False):
        """
        Compare this version with ``other`` returning -1, 0, or 1 if the
        other version is larger, the same, or smaller than this
        one. Attempts to compare to something that's not a
        ``GemVersion raises an exception.

        The comparison results have the same semantics as the legacy "cmp()"
        built-in function.
        """
        if trace:
            print(f"\nComparing: {self!r} with {other!r}")
        if isinstance(other, str):
            other = GemVersion(other)
            if trace:
                print(f"  Converted to GemVersion: {other!r}")

        if not isinstance(other, GemVersion):
            if trace:
                print(f"  Not a GemVersion: {other!r}")
            return

        if self.version == other.version:
            return 0

        lhsegments = self.canonical_segments
        if trace:
            print(f"  lhsegments: canonical_segments: {lhsegments!r}")

        rhsegments = other.canonical_segments
        if trace:
            print(f"  rhsegments: canonical_segments: {rhsegments!r}")

        if lhsegments == rhsegments:
            if trace:
                print(f"    lhsegments == rhsegments: returning 0")
            return 0

        lhsize = len(lhsegments)
        rhsize = len(rhsegments)
        if trace:
            print(f"  lhsize: {lhsize!r}")
        if trace:
            print(f"  rhsize: {rhsize!r}")

        if lhsize > rhsize:
            if trace:
                print(f"  lhsize > rhsize: limit = lhsize: {lhsize!r}")
            limit = lhsize
        else:
            if trace:
                print(f"  lhsize <= rhsize: limit = rhsize: {rhsize!r}")
            limit = rhsize

        limit -= 1

        i = 0

        if trace:
            print(f"  limit: {limit!r}, i: {i!r}")

        while i <= limit:
            if trace:
                print(f"    limit: {limit!r}, i: {i!r}")

            try:
                lhs = lhsegments[i]
            except IndexError:
                lhs = 0

            try:
                rhs = rhsegments[i]
            except IndexError:
                rhs = 0

            i += 1

            if trace:
                print(f"      lhs: {lhs} rhs: {rhs} i: {i!r}")

            if lhs == rhs:
                if trace:
                    print(f"      lhs == rhs: continue")
                continue

            if isinstance(lhs, str) and isinstance(rhs, int):
                if trace:
                    print(f"      isinstance(lhs, str): {type(lhs)!r}")
                    print(f"      isinstance(rhs, int): {type(rhs)!r}")
                    print(f"      return -1")
                return -1

            if isinstance(lhs, int) and isinstance(rhs, str):
                if trace:
                    print(f"      isinstance(lhs, int): {type(lhs)!r}")
                    print(f"      isinstance(rhs, str): {type(rhs)!r}")
                    print(f"      return 1")
                return 1

            result = (lhs > rhs) - (lhs < rhs)
            if trace:
                print(f"      (lhs > rhs) - (lhs < rhs):{result!r}")
                print(f"      return {result}")

            return result

        if trace:
            print(f"  all options evaluated: return 0")
        return 0


GemConstraint = namedtuple("GemConstraint", ["op", "version"])
GemConstraint.to_string = lambda gc: f"{gc.op} {gc.version}"


def sort_constraints(constraints):
    """
    Return a sorted sequence of unique GemConstraints.
    """
    constraints = sorted(constraints, key=lambda gc: (gc.version, gc.op))
    consts = []
    for gc in constraints:
        if gc in consts:
            continue
        consts.append(gc)
    return consts


def tilde_comparator(version, requirement, trace=False):
    """
    Return True if ``version`` GemVersion satisfies ``requirement`` GemVersion
    according to the Rubygems tilde semantics.
    """
    if trace:
        print(f"      tilde_comparator: version: {version!r}, requirement: {requirement!r}")
        print(f"         version >= requirement: {version >= requirement!r}")
        print()
        print(
            f"         version.release() < requirement.bump(): {version.release()!r} "
            f"< {requirement.bump()!r}: {version.release() < requirement.bump()!r}"
        )

    return version >= requirement and version.release() < requirement.bump()


class GemRequirement:
    """
    A gem requirement using the Gem notation.
    """

    equal_op = operator.eq
    comparators_by_op = {
        "=": equal_op,
        "!=": operator.ne,
        ">": operator.gt,
        "<": operator.lt,
        ">=": operator.ge,
        "<=": operator.le,
        "~>": tilde_comparator,
    }

    quoted = "|".join(re.escape(op) for op in comparators_by_op)

    PATTERN_RAW = f"\\s*({quoted})?\\s*({GemVersion.VERSION_PATTERN})\\s*"

    # A regular expression that matches a requirement
    PATTERN = re.compile(f"^{PATTERN_RAW}$")

    # The default requirement matches any version
    DEFAULT_CONSTRAINT = GemConstraint(">=", GemVersion(0))

    def __init__(self, *requirements):
        """
        Initialize a GemRequirement from a sequence of ``requirements``
        converted to a constraints sequence of GemConstraint.
        """
        if not requirements:
            self.constraints = (GemRequirement.DEFAULT_CONSTRAINT,)
        else:
            self.constraints = tuple([GemRequirement.parse(r) for r in requirements])

    def __str__(self):
        gcs = [gc.to_string() for gc in sort_constraints(self.constraints)]
        return ", ".join(gcs)

    def __repr__(self):
        gcs = ", ".join(repr(gc.to_string()) for gc in sort_constraints(self.constraints))
        return f"GemRequirement({gcs})"

    @classmethod
    def from_string(cls, requirements):
        """
        Return a GemRequirement build from a lockfile-style ``requirements``
        string.

        For example::
        >>> gr1 = GemRequirement(">= 1.0.1", "~> 1.0")
        >>> gr2 = GemRequirement.from_string(" (>= 1.0.1, ~> 1.0)")
        >>> assert gr1 == gr2, (gr1, gr2)
        """
        reqs = requirements.strip().strip("()")
        reqs = [r.strip() for r in reqs.split(",")]
        return cls(*reqs)

    def for_lockfile(self):
        """
        Return a string representing this list of requirements suitable for use
        in a lockfile.

        For example::
        >>> gr = GemRequirement(">= 1.0.1", "~> 1.0")
        >>> gf_flf = gr.for_lockfile()
        >>> assert gf_flf == " (~> 1.0, >= 1.0.1)", gf_flf
        """
        gcs = [gc.to_string() for gc in sort_constraints(self.constraints)]
        gcs = ", ".join(gcs)
        return f" ({gcs})"

    def dedupe(self):
        """
        Return a new GemRequirement with sorted and unique constraints.
        """
        return GemRequirement(*sort_constraints(self.constraints))

    def simplify(self):
        """
        Return a new simplified GemRequirement with:
        - sorted and unique constraints.
        - where ~> constraints are replaced by simpler contrainsts.
        """
        constraints = []
        for const in self.constraints:
            if const.op == "~>":
                low_high = get_tilde_constraints(const)
                constraints.extend(low_high)
            else:
                constraints.append(const)
        return GemRequirement(*sort_constraints(constraints))

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        # An == check is always necessary
        if sort_constraints(self.constraints) == sort_constraints(other.constraints):
            stilde = self.tilde_requirements()
            if not stilde:
                # An == check is sufficient unless any requirements use ~>
                return True
            else:
                # If any requirements use ~> we use the stricter `#eql?` that
                # also checks that version precision is the same
                otilde = other.tilde_requirements()
                if len(stilde) != len(otilde):
                    return False
                for st, ot in zip(stilde, otilde):
                    if st.op != ot.op or not st.version.equal_strictly(ot.version):
                        return False
                return True
        return False

    def exact(self):
        """
        Return True if the requirement is for only an exact version.

        For example:
        >>> GemRequirement(">= 1.0.1", "~> 1.0").exact()
        False
        >>> GemRequirement("= 1.0.1", "~> 1.0").exact()
        False
        >>> GemRequirement("= 1.0.1").exact()
        True
        """
        return len(self.constraints) == 1 and self.constraints[0].op == "="

    @classmethod
    def create(cls, reqs):
        """
        Return a GemRequirement built from a single requirement string or a list
        of requirement strings.
        """
        if isinstance(reqs, list):
            return cls(*reqs)
        else:
            return cls(reqs)

    @classmethod
    def parse(cls, requirement):
        """
        Return a GemConstraint tuple of (operator string, GemVersion object)
        parsed from a single ``requirement`` string such as "> 3.0". Also
        accepts a two-tuple or list of ("op", "version") or a single GemVersion or a
        GemConstraint).
        """
        if isinstance(requirement, GemVersion):
            return GemConstraint("=", requirement)

        if isinstance(requirement, (tuple, list, GemConstraint)):
            return GemConstraint(*requirement)

        if not isinstance(requirement, str):
            raise InvalidRequirementError("Illformed requirement {requirement!r}")

        match = cls.PATTERN.match(str(requirement))
        if not match:
            raise InvalidRequirementError("Illformed requirement {requirement!r}")

        if match.group(1) == ">=" and match.group(2) == "0":
            return cls.DEFAULT_CONSTRAINT
        else:
            op = match.group(1) if match.group(1) else "="
            return GemConstraint(op, GemVersion(match.group(2)))

    def satisfied_by(self, version, trace=False):
        """
        Return True if the ``version`` GemVersion or version string or int
        satisfies all the constraints of this requirement. Raise an
        InvalidVersionError with an invalid ``version``.
        """
        if trace:
            print(f"\nis {self!r} satisfied_by: {version!r} ?")
        if not isinstance(version, GemVersion):
            version = GemVersion(version)
            if trace:
                print(f" converting version to GemVersion: {version!r}")

        if not self.constraints:
            raise InvalidRequirementError(self)

        for constraint in self.constraints:
            if trace:
                print(f"  processing: {constraint!r}")

            op = constraint.op
            comparator = self.comparators_by_op[op]
            if trace:
                print(f"    got comparator: {comparator!r}")
            satisfying = comparator(version, constraint.version)
            if trace:
                print(f"    {self!r} is satisfied by: {version!r}: {satisfying!r}")
                print(f"    {version!r} {op} {constraint.version!r}: {satisfying!r}")
            if not satisfying:
                return False

        return True

    def tilde_requirements(self):
        """
        Return a sorted sequence of all pessimistic "~>" GemConstraint.
        """
        constraints = sort_constraints(self.constraints)
        return [gc for gc in constraints if gc.op == "~>"]


def get_tilde_constraints(constraint):
    """
    Return a tuple of two GemConstraint representing the lower and upper
    bound of a version range ``string`` that uses a tilde "~>" pessimistic operator.
    Raise a ValueError if this is not a tilde range.

    For example:
    >>> lower_bound, upper_bound = get_tilde_constraints(GemConstraint("~>", GemVersion("1.0.2")))
    >>> vlow = GemVersion("1.0.2")
    >>> vup = GemVersion("1.1.0")
    >>> assert lower_bound == GemConstraint(op=">=", version=vlow)
    >>> assert upper_bound == GemConstraint(op="<", version=vup)
    """
    if not isinstance(constraint, GemConstraint) or not constraint.op == "~>":
        raise ValueError(f"Invalid tilde GemConstraint: {constraint!r}")
    version = constraint.version
    assert isinstance(version, GemVersion)
    lower_bound = version.release()
    upper_bound = lower_bound.bump()

    return (
        GemConstraint(op=">=", version=lower_bound),
        GemConstraint(op="<", version=upper_bound),
    )



#
# Copyright (c) Christoph Reiter
# SPDX-License-Identifier: MIT
# Version utility extracted from msys2 https://github.com/msys2/msys2-web/
# and further stripped down.
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

import re
from itertools import zip_longest
from typing import Dict
from typing import List
from typing import Optional
from typing import Set
from typing import Tuple

from univers.utils import cmp


def vercmp(v1: str, v2: str) -> int:
    def split(v: str) -> Tuple[str, str, Optional[str]]:
        if ":" in v:
            e, v = v.split(":", 1)
        else:
            e, v = ("0", v)

        r: Optional[str] = None
        if "-" in v:
            v, r = v.rsplit("-", 1)
        else:
            v, r = (v, None)

        return (e, v, r)

    digit, alpha, other = range(3)

    def get_type(c: str) -> int:
        assert c
        if c.isdigit():
            return digit
        elif c.isalpha():
            return alpha
        else:
            return other

    def parse(v: str) -> List[str]:
        parts: List[str] = []
        current = ""
        for c in v:
            if not current:
                current += c
            else:
                if get_type(c) == get_type(current):
                    current += c
                else:
                    parts.append(current)
                    current = c

        if current:
            parts.append(current)

        return parts

    def rpmvercmp(v1: str, v2: str) -> int:
        for p1, p2 in zip_longest(parse(v1), parse(v2), fillvalue=None):
            if p1 is None:
                if get_type(p2) == alpha:
                    return 1
                return -1
            elif p2 is None:
                if get_type(p1) == alpha:
                    return -1
                return 1

            t1 = get_type(p1)
            t2 = get_type(p2)
            if t1 != t2:
                if t1 == digit:
                    return 1
                elif t2 == digit:
                    return -1
                elif t1 == other:
                    return 1
                elif t2 == other:
                    return -1
            elif t1 == other:
                ret = cmp(len(p1), len(p2))
                if ret != 0:
                    return ret
            elif t1 == digit:
                ret = cmp(int(p1), int(p2))
                if ret != 0:
                    return ret
            elif t1 == alpha:
                ret = cmp(p1, p2)
                if ret != 0:
                    return ret

        return 0

    e1, v1, r1 = split(v1)
    e2, v2, r2 = split(v2)

    ret = rpmvercmp(e1, e2)
    if ret == 0:
        ret = rpmvercmp(v1, v2)
        if ret == 0 and r1 is not None and r2 is not None:
            ret = rpmvercmp(r1, r2)

    return ret


def extract_upstream_version(version: str) -> str:
    return version.rsplit("-")[0].split("+", 1)[0].split("~", 1)[-1].split(":", 1)[-1]


def strip_vcs(package_name: str) -> str:
    if package_name.endswith(("-cvs", "-svn", "-hg", "-darcs", "-bzr", "-git")):
        return package_name.rsplit("-", 1)[0]
    return package_name


def arch_version_to_msys(v: str) -> str:
    return v.replace(":", "~")


def version_is_newer_than(v1: str, v2: str) -> bool:
    return vercmp(v1, v2) == 1


def split_depends(deps: List[str]) -> Dict[str, Set[str]]:
    r: Dict[str, Set[str]] = {}
    for d in deps:
        parts = re.split("([<>=]+)", d, 1)
        first = parts[0].strip()
        second = "".join(parts[1:]).strip()
        r.setdefault(first, set()).add(second)
    return r


def split_optdepends(deps: List[str]) -> Dict[str, Set[str]]:
    r: Dict[str, Set[str]] = {}
    for d in deps:
        if ":" in d:
            a, b = d.split(":", 1)
            a, b = a.strip(), b.strip()
        else:
            a, b = d.strip(), ""
        e = r.setdefault(a, set())
        if b:
            e.add(b)
    return r



#
# Copyright [2017] The Climate Corporation (https://climate.com)
# https://github.com/TheClimateCorporation/python-dpkg
# original author: Nathan J. Meh

# Copyright (c) nexB Inc. and others.
# http://nexb.com and https://github.com/nexB/debian_inspector/

# SPDX-License-Identifier: Apache-2.0
# this has been significantly modified from the original
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

from unittest import TestCase

import pytest

import univers.debian as version
from univers.debian import compare_strings
from univers.debian import compare_versions
from univers.debian import Version

"""
Parse, compare and sort Debian package versions.

This has been substantially modified from python-dpkg to extract the version
comparison code.
"""


def get_non_digit_prefix(s):
    val = version.get_non_digit_prefix(list(s))
    return "".join(val)


def get_digit_prefix(s):
    return version.get_digit_prefix(list(s))


class DebianVersionTest(TestCase):
    def test_Version_from_string_epoch(self):
        assert Version.from_string("0").epoch == 0
        assert Version.from_string("0:0").epoch == 0
        assert Version.from_string("1:0").epoch == 1

    def test_Version_from_string_validates(self):
        self.assertRaises(ValueError, Version.from_string, "a")
        self.assertRaises(ValueError, Version.from_string, "1a:0")
        self.assertRaises(ValueError, Version.from_string, "0:a.0.0-foo")

    def test_Version_from_string_tuples(self):
        assert Version.from_string("00").tuple() == (0, "00", "0")
        assert Version.from_string("00-00").tuple() == (0, "00", "00")
        assert Version.from_string("0:0").tuple() == (0, "0", "0")
        assert Version.from_string("0:0-0").tuple() == (0, "0", "0")
        assert Version.from_string("0:0.0").tuple() == (0, "0.0", "0")
        assert Version.from_string("0:0.0-0").tuple() == (0, "0.0", "0")
        assert Version.from_string("0:0.0-00").tuple() == (0, "0.0", "00")

    def test_get_non_digit_prefix(self):
        assert get_non_digit_prefix("") == ""
        assert get_non_digit_prefix("0") == ""
        assert get_non_digit_prefix("00") == ""
        assert get_non_digit_prefix("0a") == ""
        assert get_non_digit_prefix("a") == "a"
        assert get_non_digit_prefix("a0") == "a"
        assert get_non_digit_prefix("aHAD0030") == "aHAD"

    def test_get_digit_prefix(self):
        assert get_digit_prefix("00") == 0
        assert get_digit_prefix("0") == 0
        assert get_digit_prefix("0a") == 0
        assert get_digit_prefix("12a") == 12
        assert get_digit_prefix("a") == 0
        assert get_digit_prefix("a0") == 0
        assert get_digit_prefix("arttr23123") == 0
        assert get_digit_prefix("123sdf") == 123
        assert get_digit_prefix("0123sdf") == 123

    def test_compare_strings(self):
        assert compare_strings("~", ".") == -1
        assert compare_strings("~", "a") == -1
        assert compare_strings("a", ".") == -1
        assert compare_strings("a", "~") == 1
        assert compare_strings(".", "~") == 1
        assert compare_strings(".", "a") == 1
        assert compare_strings(".", ".") == 0
        assert compare_strings("0", "0") == 0
        assert compare_strings("a", "a") == 0

    def test_compare_strings_can_sort(self):
        # taken from
        # http://www.debian.org/doc/debian-policy/ch-controlfields.html#s-f-Version
        result = sorted(["a", "", "~", "~~a", "~~"], key=version.compare_strings_key)
        expected = ["~~", "~~a", "~", "", "a"]
        assert expected == result

    def test_compare_strings_more(self):
        # note that these are testing a single revision string, not the full
        # upstream+debian version.  IOW, "0.0.9-foo" is an upstream or debian
        # revision onto itself, not an upstream of 0.0.9 and a debian of foo.

        # equals
        assert compare_strings("0", "0") == 0
        assert compare_strings("0", "00") == 0
        assert compare_strings("00.0.9", "0.0.9") == 0
        assert compare_strings("0.00.9-foo", "0.0.9-foo") == 0
        assert compare_strings("0.0.9-1.00foo", "0.0.9-1.0foo") == 0

        # less than
        assert compare_strings("0.0.9", "0.0.10") == -1
        assert compare_strings("0.0.9-foo", "0.0.10-foo") == -1
        assert compare_strings("0.0.9-foo", "0.0.10-goo") == -1
        assert compare_strings("0.0.9-foo", "0.0.9-goo") == -1
        assert compare_strings("0.0.10-foo", "0.0.10-goo") == -1
        assert compare_strings("0.0.9-1.0foo", "0.0.9-1.1foo") == -1

        # greater than
        assert compare_strings("0.0.10", "0.0.9") == 1
        assert compare_strings("0.0.10-foo", "0.0.9-foo") == 1
        assert compare_strings("0.0.10-foo", "0.0.9-goo") == 1
        assert compare_strings("0.0.9-1.0foo", "0.0.9-1.0bar") == 1

    def test_compare_versions(self):
        # "This [the epoch] is a single (generally small) unsigned integer.
        # It may be omitted, in which case zero is assumed."
        assert compare_versions("0.0.0", "0:0.0.0") == 0
        assert compare_versions("0:0.0.0-foo", "0.0.0-foo") == 0
        assert compare_versions("0.0.0-a", "0:0.0.0-a") == 0

        # "The absence of a debian_revision is equivalent to a debian_revision of 0."
        assert compare_versions("0.0.0", "0.0.0-0") == 0
        # tricksy:
        assert compare_versions("0.0.0", "0.0.0-00") == 0

        # combining the above
        assert compare_versions("0.0.0-0", "0:0.0.0") == 0

        # explicitly equal
        assert compare_versions("0.0.0", "0.0.0") == 0
        assert compare_versions("1:0.0.0", "1:0.0.0") == 0
        assert compare_versions("0.0.0-10", "0.0.0-10") == 0
        assert compare_versions("2:0.0.0-1", "2:0.0.0-1") == 0

        # less than
        assert compare_versions("0.0.0-0", "0:0.0.1") == -1
        assert compare_versions("0.0.0-0", "0:0.0.0-a") == -1
        assert compare_versions("0.0.0-0", "0:0.0.0-1") == -1
        assert compare_versions("0.0.9", "0.0.10") == -1
        assert compare_versions("0.9.0", "0.10.0") == -1
        assert compare_versions("9.0.0", "10.0.0") == -1

        assert compare_versions("1.2.3-1~deb7u1", "1.2.3-1") == -1
        assert compare_versions("2.7.4+reloaded2-13ubuntu1", "2.7.4+reloaded2-13+deb9u1") == -1
        assert compare_versions("2.7.4+reloaded2-13", "2.7.4+reloaded2-13+deb9u1") == -1

        # greater than
        assert compare_versions("0.0.1-0", "0:0.0.0") == 1
        assert compare_versions("0.0.0-a", "0:0.0.0-1") == 1
        assert compare_versions("0.0.0-a", "0:0.0.0-0") == 1
        assert compare_versions("0.0.9", "0.0.1") == 1
        assert compare_versions("0.9.0", "0.1.0") == 1
        assert compare_versions("9.0.0", "1.0.0") == 1

        assert compare_versions("1.2.3-1", "1.2.3-1~deb7u1") == 1
        assert compare_versions("2.7.4+reloaded2-13+deb9u1", "2.7.4+reloaded2-13ubuntu1") == 1
        assert compare_versions("2.7.4+reloaded2-13+deb9u1", "2.7.4+reloaded2-13") == 1

        # unicode
        assert compare_versions("2:0.0.44-1", "2:0.0.44-nobin") == -1
        assert compare_versions("2:0.0.44-nobin", "2:0.0.44-1") == 1
        assert compare_versions("2:0.0.44-1", "2:0.0.44-1") == 0

    @pytest.mark.xfail(reason="Not yet supported")
    def test_can_parse_complex_version_is_not_invalid(self):
        Version.from_string("1:1.12_1.12.6-1+deb9u1build0.18.04.1")



#
# Copyright (c) nexB Inc. and others
# Copyright (c) SAS Institute Inc.
# SPDX-License-Identifier: Apache-2.0
# this has been significantly modified from the original
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

import io
import os
import re
import unittest

from univers import rpm as vercmp

get_vercmp_test = re.compile(
    r"(.*)" r"RPMVERCMP\(" r"([^, ]*) *, *" r"([^, ]*) *, *" r"([^\)]*)" r"\).*"
).match


def parse_rpmvercmp_tests(fobj, with_buggy_comparisons=True):
    """
    Yield triples (version1, version2, comparison result) describing the RPM
    version comparison tests found in the rpmvercmp.at test file-like  object at
    ``fobj`` Optionally include "buggy" upstream tests.
    """
    for line in fobj:
        line = line.strip()
        if not line:
            continue
        if not with_buggy_comparisons and line.startswith("dnl"):
            continue
        match = get_vercmp_test(line)
        if not match:
            continue
        yield match.group(2), match.group(3), int(match.group(4))


class ParserTest(unittest.TestCase):
    def test_parse_without_buggy(self):
        fobj = io.StringIO(
            """
bla
bla
RPMVERCMP(1, 2, -1)
dnl RPMVERCMP(1a, 1b, -1)
"""
        )
        parser = parse_rpmvercmp_tests(fobj, with_buggy_comparisons=False)
        expected = [("1", "2", -1)]
        assert list(parser) == expected

    def test_parse_with_buggy(self):
        fobj = io.StringIO(
            """
bla
bla
RPMVERCMP(1, 2, -1)
dnl RPMVERCMP(1a, 1b, -1)
"""
        )
        parser = parse_rpmvercmp_tests(fobj, with_buggy_comparisons=True)
        expected = [("1", "2", -1), ("1a", "1b", -1)]
        assert list(parser) == expected


class VersionCompareTest(unittest.TestCase):
    pass


def create_test_function(ver1, ver2, expected, name):
    """
    Return a test function closed on test arguments.
    """

    def test_rpm_version(self):
        print(f"testing (ver1={ver1}, ver2={ver2}, expected=expected{expected}")
        result = vercmp.vercmp(ver1, ver2)
        if result != expected:
            assert result == (expected, ver1, ver2)

    test_rpm_version.__name__ = name
    return test_rpm_version


def get_tests():
    """
    Yield test function from rpmvercmp.at data.
    """
    test_file = os.path.join(os.path.dirname(__file__), "data", "rpmvercmp.at")

    with io.open(test_file, encoding="utf-8") as rpmtests:
        tests = list(parse_rpmvercmp_tests(rpmtests, with_buggy_comparisons=True))
        for test_count, (ver1, ver2, expected) in enumerate(tests, 1):
            name = f"test_rpm_version_{test_count}"
            yield create_test_function(ver1, ver2, expected, name)

    # Make sure we still test something, in case the m4 file drops
    # content this will fail the test
    assert test_count > 20


def build_tests():
    """
    Create tests from rpmvercmp.at data and attach them to a test class.
    """
    for i, test_func in enumerate(get_tests()):
        name = f"test_rpm_version_{i}"
        setattr(VersionCompareTest, name, test_func)


build_tests()



#
# Copyright (c) Chad Fowler, Rich Kilmer, Jim Weirich and others.
# Portions copyright (c) Engine Yard and Andre ArkoFacebook, Inc. and its affiliates.
#
# SPDX-License-Identifier: MIT
#
# Originally from https://github.com/rubygems/rubygems

from univers.gem import GemRequirement
from univers.gem import GemVersion
from univers.gem import InvalidVersionError


def assert_bumped_version_equal(expected, unbumped):
    # Assert that bumping the +unbumped+ version yields the +expected+.

    assert_version_eql(expected, GemVersion(unbumped).bump())


def test_bump():
    assert_bumped_version_equal("5.3", "5.2.4")


def test_bump_alpha():
    assert_bumped_version_equal("5.3", "5.2.4.a")


def test_bump_alphanumeric():
    assert_bumped_version_equal("5.3", "5.2.4.a10")


def test_bump_trailing_zeros():
    assert_bumped_version_equal("5.1", "5.0.0")


def test_bump_one_level():
    assert_bumped_version_equal("6", "5")


def test_eql_is_same():
    assert_version_eql("1.2", "1.2")
    assert_version_strict_equal("1.2", "1.2")

    refute_version_eql("1.2", "1.3")
    refute_version_strict_equal("1.2", "1.3")

    refute_version_strict_equal("1.2", "1.2.0")
    assert_version_eql("1.2", "1.2.0")

    assert_version_eql("1.2.b1", "1.2.b.1")
    refute_version_strict_equal("1.2.b1", "1.2.b.1")

    refute_version_strict_equal("1.2.pre.1", "1.2.0.pre.1.0")
    assert_version_eql("1.2.pre.1", "1.2.0.pre.1.0")


def test_initialize():
    for good in ["1.0", "1.0 ", " 1.0 ", "1.0\n", "\n1.0\n", "1.0"]:
        assert_version_eql("1.0", good)

    assert_version_eql("1", 1)


def test_initialize_invalid():
    invalid_versions = [
        "whatever",
        "junk",
        "1.0\n2.0" "1..2",
        "1.2\ 3.4",
    ]

    # DON'T TOUCH THIS WITHOUT CHECKING CVE-2013-4287
    invalid_versions += ["2.3422222.222.222222222.22222.ads0as.dasd0.ddd2222.2.qd3e."]

    for invalid in invalid_versions:
        try:
            GemVersion(invalid)
            raise Exception(f"exception not raised for: {invalid!r}")
        except InvalidVersionError:
            pass


def test_empty_version():
    assert GemVersion("").version == "0"
    assert GemVersion("  ").version == "0"
    assert GemVersion(" ").version == "0"


def test_prerelease():
    assert_prerelease("1.2.0.a")
    assert_prerelease("2.9.b")
    assert_prerelease("22.1.50.0.d")
    assert_prerelease("1.2.d.42")

    assert_prerelease("1.A")

    assert_prerelease("1-1")
    assert_prerelease("1-a")

    refute_prerelease("1.2.0")
    refute_prerelease("2.9")
    refute_prerelease("22.1.50.0")


def test_release():
    assert_release_equal("1.2.0", "1.2.0.a")
    assert_release_equal("1.1", "1.1.rc10")
    assert_release_equal("1.9.3", "1.9.3.alpha.5")
    assert_release_equal("1.9.3", "1.9.3")


def test_spaceship_cmp():
    def cmp(a, b):
        return a.__cmp__(b)

    # Ruby spaceship  <=>  is the same as Python legacy cmp()
    assert cmp(GemVersion("1.0"), GemVersion("1.0.0")) == 0
    assert cmp(GemVersion("1.0"), GemVersion("1.0.a")) == 1
    assert cmp(GemVersion("1.8.2"), GemVersion("0.0.0")) == 1
    assert cmp(GemVersion("1.8.2"), GemVersion("1.8.2.a")) == 1
    assert cmp(GemVersion("1.8.2.b"), GemVersion("1.8.2.a")) == 1
    assert cmp(GemVersion("1.8.2.a"), GemVersion("1.8.2")) == -1
    assert cmp(GemVersion("1.8.2.a10"), GemVersion("1.8.2.a9")) == 1
    assert cmp(GemVersion(""), GemVersion("0")) == 0
    assert cmp(GemVersion("0.beta.1"), GemVersion("0.0.beta.1")) == 0
    assert cmp(GemVersion("0.0.beta"), GemVersion("0.0.beta.1")) == -1
    assert cmp(GemVersion("0.0.beta"), GemVersion("0.beta.1")) == -1
    assert cmp(GemVersion("5.a"), GemVersion("5.0.0.rc2")) == -1
    assert cmp(GemVersion("5.x"), GemVersion("5.0.0.rc2")) == 1


def assert_version_satisfies_requirement(requirement, version):
    # Assert that +version+ satisfies the "approximate" ~> +requirement+.
    req = GemRequirement.create(requirement)
    ver = GemVersion(version)
    assert req.satisfied_by(ver)


def test_satisfies_requirement():
    assert_version_satisfies_requirement("~> 1.0", "1")
    assert_version_satisfies_requirement("~> 1.0", "1.0")
    assert_version_satisfies_requirement("~> 1.2", "1.2")
    assert_version_satisfies_requirement("~> 1.2", "1.2.0")
    assert_version_satisfies_requirement("~> 1.2", "1.2.3")
    assert_version_satisfies_requirement("~> 1.2.a", "1.2.3.a.4")
    assert_version_satisfies_requirement("~> 1.9.a", "1.9.0.dev")


def test_to_s():
    assert GemVersion("5.2.4").to_string() == "5.2.4"


def test_compare():
    assert GemVersion("0.0.1.0") > GemVersion("0.0.0.1")
    assert not GemVersion("0.0.1.0") < GemVersion("0.0.0.1")
    assert GemVersion("0.0.1.0") >= GemVersion("0.0.0.1")
    assert not GemVersion("0.0.1.0") <= GemVersion("0.0.0.1")


def test_semver():
    assert_less_than("1.0.0-alpha", "1.0.0-alpha.1")
    assert_less_than("1.0.0-alpha.1", "1.0.0-beta.2")
    assert_less_than("1.0.0-beta.2", "1.0.0-beta.11")
    assert_less_than("1.0.0-beta.11", "1.0.0-rc.1")
    assert_less_than("1.0.0-rc1", "1.0.0")
    assert_less_than("1.0.0-1", "1")


def test_segments():
    # modifying the segments of a version should not affect the segments of the cached version object
    ver = GemVersion("9.8.7")
    secondseg = ver.segments[2]
    secondseg += 1

    refute_version_eql("9.8.8", "9.8.7")
    assert GemVersion("9.8.7").segments == [9, 8, 7]


def test_split_segments():
    assert GemVersion("3.2.4-2").split_segments() == ([3, 2, 4], ["pre", 2])


def test_canonical_segments():
    assert GemVersion("1.0.0").canonical_segments == [1]
    assert GemVersion("1.0.0.a.1.0").canonical_segments == [1, "a", 1]
    assert GemVersion("1.2.3-1").canonical_segments == [1, 2, 3, "pre", 1]


def test_frozen_version():
    ver = GemVersion("1.test")
    assert_less_than(ver, GemVersion("1"))
    assert_version_eql(GemVersion("1"), ver.release())
    assert_version_eql(GemVersion("2"), ver.bump())


def assert_prerelease(version):
    # Asserts that +version+ is a prerelease.
    assert GemVersion(version).prerelease(), "#{version} is a prerelease"


def assert_release_equal(release, version):
    # Assert that +release+ is the correct non-prerelease +version+.
    assert_version_eql(release, GemVersion(version).release())


def assert_version_eql(first, second):
    # Assert that two versions are eql?. Checks both directions.
    first = GemVersion(first)
    second = GemVersion(second)
    assert first is not second
    assert first == second
    assert second == first


def refute_version_eql(first, second):
    # Refute the assumption that two versions are eql?. Checks both
    # directions.
    first = GemVersion(first)
    second = GemVersion(second)
    assert first is not second
    assert first != second
    assert second != first


def assert_version_strict_equal(first, second):
    # Assert that two versions are strictly equal
    first = GemVersion(first)
    second = GemVersion(second)
    assert first is not second
    assert first.equal_strictly(second)
    assert second.equal_strictly(first)


def refute_version_strict_equal(first, second):
    first = GemVersion(first)
    second = GemVersion(second)
    assert first is not second
    assert not first.equal_strictly(second)
    assert not second.equal_strictly(first)


def assert_less_than(left, right):
    assert GemVersion(left) < GemVersion(right)


def refute_prerelease(version):
    # Refute the assumption that +version+ is a prerelease.
    assert not GemVersion(version).prerelease()



# Copyright (c) Center for Information Technology, http://coi.gov.pl
# SPDX-License-Identifier: Apache-2.0
# this has been significantly modified from the original
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

# notes: This has been substantially modified and enhanced from the original
# puppeteer code to extract the Ruby version handling code.

from univers.gem import GemRequirement
from univers.gem import GemVersion


def test_gem_version_release():
    assert GemVersion("1.2.4.beta").release() == GemVersion("1.2.4")


def test_gem_version_bump():
    assert GemVersion("1.2.4").bump() == GemVersion("1.3.0")


def test_gem_version_compare():
    assert GemVersion("1.3.0") == GemVersion("1.3")
    assert GemVersion("1.3.0") <= GemVersion("1.3")
    assert GemVersion("1.1.3") <= GemVersion("1.3")
    assert GemVersion("1.4.pre") >= GemVersion("1.3")
    assert GemVersion("1.4.pre") != GemVersion("1.4")


def test_gem_requirement():
    assert GemRequirement("3.4").satisfied_by("3.4.0")
    assert GemRequirement("~> 3.4").satisfied_by("3.4.8")
    assert GemRequirement(">= 3.4").satisfied_by("4.4.8")
    assert GemRequirement(">= 3.4", "<4").satisfied_by("3.45.8")


def test_gem_requirement_fails1():
    assert GemRequirement(">= 3.4", "<4").satisfied_by("4.1") is False


def test_gem_requirement_fails2():
    assert GemRequirement("~> 3").satisfied_by("4.1.0.pre") is False



#
# Copyright (c) nexB Inc. and others.
# SPDX-License-Identifier: Apache-2.0
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.


def remove_spaces(string):
    return "".join(string.split())


def cmp(x, y):
    """
    Replacement for built-in Python 2 function cmp that was removed in Python 3
    From https://docs.python.org/2/library/functions.html?highlight=cmp#cmp :

        Compare the two objects x and y and return an integer according to the
        outcome. The return value is negative if x < y, zero if x == y and
        strictly positive if x > y.
    """
    if x == y:
        return 0
    elif x is None:
        return -1
    elif y is None:
        return 1
    else:
        # note that this is the minimal replacement function
        return (x > y) - (x < y)



#
# Copyright (c) Peter Odding <peter@peterodding.com>
# URL: https://github.com/xolox/python-deb-pkg-tools
# SPDX-License-Identifier: MIT
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.


from unittest import TestCase

from univers.debian import Version


class DebPkgToolsTestCase(TestCase):
    def test_version_sorting(self):
        # Check version sorting implemented on top of `=' and `<<' comparisons.
        expected_order = ["0.1", "0.5", "1.0", "2.0", "3.0", "1:0.4", "2:0.3"]
        assert list(sorted(expected_order)) != expected_order
        result = [str(v) for v in sorted(map(Version.from_string, expected_order))]
        assert result == expected_order

    def test_version_comparison(self):
        assert Version.from_string("1.0") > Version.from_string("0.5")
        assert Version.from_string("1:0.5") > Version.from_string("2.0")
        assert not Version.from_string("0.5") > Version.from_string("2.0")

        assert Version.from_string("0.75") >= Version.from_string("0.5")
        assert Version.from_string("0.50") >= Version.from_string("0.5")
        assert Version.from_string("1:0.5") >= Version.from_string("5.0")
        assert not Version.from_string("0.2") >= Version.from_string("0.5")

        assert Version.from_string("0.5") < Version.from_string("1.0")
        assert Version.from_string("2.0") < Version.from_string("1:0.5")
        assert not Version.from_string("2.0") < Version.from_string("0.5")

        assert Version.from_string("0.5") <= Version.from_string("0.75")
        assert Version.from_string("0.5") <= Version.from_string("0.50")
        assert Version.from_string("5.0") <= Version.from_string("1:0.5")
        assert not Version.from_string("0.5") <= Version.from_string("0.2")

        assert Version.from_string("42") == Version.from_string("42")
        assert Version.from_string("0.5") == Version.from_string("0:0.5")
        assert not Version.from_string("0.5") == Version.from_string("1.0")

        assert Version.from_string("1") != Version.from_string("0")
        assert not Version.from_string("0.5") != Version.from_string("0:0.5")

        assert Version.from_string("1.3~rc2") < Version.from_string("1.3")



# -*- coding: utf-8 -*-
#
# Copyright (c) nexB Inc. and others.
# SPDX-License-Identifier: Apache-2.0
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

import json
import os
import re
import unittest

from univers.version_range import VersionRange
from unittest.case import expectedFailure


def create_test_function(
    description,
    vers,
    canonical_vers,
    is_invalid,
    scheme,
    constraints,
    test_func_prefix="test_vers_",
    **kwargs
):
    """
    Return a new (test function, test_name) where the test_function closed on
    test arguments. If is_error is True the tests are expected to raise an
    Exception.
    """
    if is_invalid:

        def test_vers(self):
            try:
                VersionRange.from_string(vers)
                self.fail("Should raise a ValueError")
            except ValueError:
                pass

            try:
                VersionRange.from_string(canonical_vers)
                self.fail("Should raise a ValueError")
            except ValueError:
                pass

    else:

        def test_vers(self):
            # parsing the test canonical `vers` then re-building a `vers` from these
            # parsed components should return the test canonical `vers`
            cano = VersionRange.from_string(vers)
            assert canonical_vers == cano.to_string()

            # parsing the test `vers` should return the components parsed from the
            # test canonical `vers`
            parsed = VersionRange.from_string(canonical_vers)
            assert str(cano) == str(parsed)

            # parsing the test `vers` then re-building a `vers` from these parsed
            # components should return the test canonical `vers`
            assert canonical_vers == parsed.to_string()

            # building a `vers` from the test ranges should return the test
            # canonical `vers`
            built = VersionRange(scheme, constraints)
            assert canonical_vers == built.to_string()

    # create a good function name for use in test discovery
    if not description:
        description = vers
    if is_invalid:
        test_func_prefix += "is_invalid_"
    test_name = python_safe_name(test_func_prefix + description)
    test_vers.__name__ = test_name
    test_vers.funcname = test_name
    return test_vers, test_name


def python_safe_name(s):
    """
    Return a name derived from string `s` safe to use as a Python function name.

    For example:
    >>> s = "not `\\a /`good` -safe name ??"
    >>> assert python_safe_name(s) == 'not_good_safe_name'
    """
    no_punctuation = re.compile(r"[\W_]", re.MULTILINE).sub
    s = s.lower()
    s = no_punctuation(" ", s)
    s = "_".join(s.split())
    return s


class VersTest(unittest.TestCase):
    pass


def build_tests(clazz=VersTest, test_file="test-suite-data.json"):
    """
    Dynamically build test methods for each vers test found in the `test_file`
    JSON file and attach a test method to the `clazz` class.
    """
    test_data_dir = os.path.join(os.path.dirname(__file__), "data")
    test_file = os.path.join(test_data_dir, test_file)

    with open(test_file) as tf:
        tests_data = json.load(tf)
    for items in tests_data:
        test_func, test_name = create_test_function(**items)
        # TODO: remove once implemented
        test_func = expectedFailure(test_func)
        # attach that method to the class
        setattr(clazz, test_name, test_func)


# build_tests()



#
# Copyright (c) nexB Inc. and others.
# SPDX-License-Identifier: Apache-2.0
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

import subprocess
import unittest


class BaseTests(unittest.TestCase):
    def test_codestyle(self):
        args = "venv/bin/black --check -l 100 setup.py src tests"
        try:
            subprocess.check_output(args.split())
        except subprocess.CalledProcessError as e:
            print("===========================================================")
            print(e.output)
            print("===========================================================")
            raise Exception(
                "Black style check failed; please format the code using:\n"
                "  python -m black -l 100 setup.py src tests",
                e.output,
            ) from e



#
# Copyright (c) nexB Inc. and others.
# SPDX-License-Identifier: Apache-2.0
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

from packaging import version as packaging_version
from unittest import TestCase

from univers import versions

# version comparison is already tested at:
# https://github.com/pypa/packaging/blob/main/tests/test_version.py


class TestPYPIVersion(TestCase):
    def test_constructor(self):
        pypi_version = versions.PypiVersion("2.4.5")
        assert pypi_version.value == packaging_version.Version("2.4.5")
        self.assertRaises(versions.InvalidVersion, versions.PypiVersion, "2.//////")



#
# Copyright (c) SAS Institute Inc.
# SPDX-License-Identifier: Apache-2.0
# Version comparison utility extracted from pymaven and further stripped down
# and significantly modified from the original at pymaven
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

import functools
from itertools import zip_longest

from univers.utils import cmp


# TODO: Use common exceptions with other modules
class VersionRangeParseError(ValueError):
    pass


class RestrictionParseError(ValueError):
    pass


EXCLUSIVE_CLOSE = ")"
EXCLUSIVE_OPEN = "("
INCLUSIVE_CLOSE = "]"
INCLUSIVE_OPEN = "["

# Known qualifiers, oldest to newest
QUALIFIERS = ["alpha", "beta", "milestone", "rc", "snapshot", "", "sp"]

# Well defined aliases
ALIASES = {
    "ga": "",
    "final": "",
    "cr": "rc",
}


def list2tuple(l):
    return tuple(list2tuple(x) if isinstance(x, list) else x for x in l)


@functools.total_ordering
class Restriction(object):
    """Describes a restriction in versioning"""

    def __init__(self, spec=None):
        """Create a restriction

        Restrictions are specified using a semi-mathematical notation:

        * 1.0: "Soft" requirement on 1.0 (just a recommendation, if it matches
          all other ranges for the dependency)
        * [1.0]: "Hard" requirement on 1.0
        * (,1.0]: x <= 1.0
        * [1.2,1.3]: 1.2 <= x <= 1.3
        * [1.0,2.0): 1.0 <= x < 2.0
        * [1.5,): x >= 1.5
        * (,1.0],[1.2,): x <= 1.0 or x >= 1.2; multiple sets are comma-separated
        * (,1.1),(1.1,): this excludes 1.1 (for example if it is known not to
          work in combination with this library)

        :param str spec: Restriction specification
        """
        self.lower_bound = None
        self.upper_bound = None
        self.lower_bound_inclusive = False
        self.upper_bound_inclusive = False

        if not spec:
            return

        self.lower_bound_inclusive = spec.strip()[0] == INCLUSIVE_OPEN
        self.upper_bound_inclusive = spec.strip()[-1] == INCLUSIVE_CLOSE

        _spec = spec[1:-1].strip()
        if "," in _spec:
            lower_bound, upper_bound = _spec.split(",")
            if lower_bound and lower_bound == upper_bound:
                raise RestrictionParseError("Range cannot have identical boundaries: %s" % spec)

            self.lower_bound = Version(lower_bound) if lower_bound else None
            self.upper_bound = Version(upper_bound) if upper_bound else None

            if self.lower_bound and self.upper_bound and self.upper_bound < self.lower_bound:
                raise RestrictionParseError("Range defies version ordering: %s" % spec)
        else:
            # single version restriction
            if not self.lower_bound_inclusive or not self.upper_bound_inclusive:
                raise RestrictionParseError("Single version must be surrounded by []: %s" % spec)
            version = Version(_spec)
            self.lower_bound = version
            self.upper_bound = version

    def __contains__(self, version):
        """Return true if version is contained within the restriction

        version must be greater than the lower bound (or equal to it if the
        lower bound is inclusive) and less than the upper bound ( or equal to it
        if the upper bound is inclusive).
        """
        if self.lower_bound:
            if self.lower_bound == version and not self.lower_bound_inclusive:
                return False

            if self.lower_bound > version:
                return False

        if self.upper_bound:
            if self.upper_bound == version and not self.upper_bound_inclusive:
                return False

            if self.upper_bound < version:
                return False

        return True

    def __cmp__(self, other):
        if self is other:
            return 0

        if not isinstance(other, self.__class__):
            if isinstance(other, str):
                return cmp(self, self.__class__(other))
            return 1

        result = cmp(self.lower_bound, other.lower_bound)
        if result == 0:
            result = cmp(self.lower_bound_inclusive, other.lower_bound_inclusive)
            if result == 0:
                result = cmp(self.upper_bound, other.upper_bound)
                if result == 0:
                    result = cmp(self.upper_bound_inclusive, other.upper_bound_inclusive)
        return result

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __hash__(self):
        return hash(
            (
                self.lower_bound,
                self.lower_bound_inclusive,
                self.upper_bound,
                self.upper_bound_inclusive,
            )
        )

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def __str__(self):
        s = "{open}{lower}{comma}{upper}{close}".format(
            open=(INCLUSIVE_OPEN if self.lower_bound_inclusive else EXCLUSIVE_OPEN),
            lower=self.lower_bound if self.lower_bound is not None else "",
            comma="," if self.lower_bound != self.upper_bound else "",
            upper=(
                self.upper_bound
                if self.upper_bound is not None and self.upper_bound != self.lower_bound
                else ""
            ),
            close=(INCLUSIVE_CLOSE if self.upper_bound_inclusive else EXCLUSIVE_CLOSE),
        )
        return s

    def __repr__(self):
        return "<%s.%s(%r, %r, %r, %r)>" % (
            self.__module__,
            "Restriction",
            self.lower_bound,
            self.lower_bound_inclusive,
            self.upper_bound,
            self.upper_bound_inclusive,
        )

    @classmethod
    def fromstring(cls, spec):
        return cls(spec)


@functools.total_ordering
class VersionRange(object):
    """Version range specification

    Valid ranges are comma separated range specifications
    """

    def __init__(self, spec):
        """Create a VersionRange from a string specification

        :param spec string representation of a version or version range
        :type spec str
        :return a new VersionRange
        :rtype VersionRange
        :raises RuntimeError if the range is invalid in some way
        """
        restrictions = []
        version = None

        _spec = spec[:]
        lower_bound = None
        upper_bound = None
        while _spec.startswith(EXCLUSIVE_OPEN) or _spec.startswith(INCLUSIVE_OPEN):
            exclusive_close = _spec.find(EXCLUSIVE_CLOSE)
            inclusive_close = _spec.find(INCLUSIVE_CLOSE)

            close = inclusive_close
            if inclusive_close < 0 or 0 <= exclusive_close < inclusive_close:
                # close is exclusive
                close = exclusive_close

            if close < 0:
                raise VersionRangeParseError("Unbounded range: %s" % spec)

            restriction = Restriction(_spec[0 : close + 1])

            if lower_bound is None:
                lower_bound = restriction.lower_bound

            if upper_bound is not None:
                if restriction.lower_bound is None or restriction.lower_bound < upper_bound:
                    raise VersionRangeParseError("Ranges overlap: %s" % spec)
            restrictions.append(restriction)
            upper_bound = restriction.upper_bound

            _spec = _spec[close + 1 :]
            if _spec and _spec.startswith(","):
                # pop off leading comma
                _spec = _spec[1:]

        if _spec:
            if restrictions:
                raise VersionRangeParseError(
                    "Only fully-qualified sets allowed in multiple set" " scenario: %s" % spec
                )
            else:
                version = Version(_spec)
                # add the "everything" restriction
                restrictions.append(Restriction())

        self.version = version
        self.restrictions = tuple(restrictions)

    def __cmp__(self, other):
        if self is other:
            return 0

        if not isinstance(other, self.__class__):
            if isinstance(other, str):
                return cmp(self, self.__class__(other))
            elif isinstance(other, Version):
                return cmp(other, self)
            return 1

        result = cmp(self.version, other.version)
        if result == 0:
            result = cmp(self.restrictions, other.restrictions)

        return result

    def __contains__(self, version):
        return any((version in r) for r in self.restrictions)

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __hash__(self):
        return hash((self.version, self.restrictions))

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def __str__(self):
        if self.version:
            return str(self.version)
        else:
            return ",".join(str(r) for r in self.restrictions)

    def __repr__(self):
        return "<%s.%s(%r, %r)>" % (
            self.__module__,
            "VersionRange",
            self.version,
            self.restrictions,
        )

    def _intersection(self, l1, l2):
        """Return the intersection of l1 and l2

        :param l1 list of restrictions
        :type l1 [Restriction, ...]
        :param l2 list of restrictions
        :type l2 [Restriction, ...]
        :return Intersection of l1 and l2
        :rtype [Restriction, ...]
        """
        raise NotImplementedError

    @classmethod
    def fromstring(cls, spec):
        return cls(spec)

    @classmethod
    def from_version(cls, version):
        return cls(str(version))

    def restrict(self, version_range):
        """Returns a new VersionRange that is a restriction of this
        and the specified version range.

        Prefers this version over the specified version range

        :param version_range the version range that will restrict this range
        :type version_range VersionRange
        :return intersection of this version range and the specified one
        :rypte VersionRange
        """
        raise NotImplementedError

    def match_version(self, versions):
        matched = None
        for version in sorted(versions, reverse=True):
            if version in self:
                matched = version
                break
        return matched


@functools.total_ordering
class Version(object):
    """Maven version objecjt"""

    def __init__(self, version):
        """Create a maven version

        The version string is examined one character at a time.

        There's a buffer containing the current text - all characters are
        appended, except for '.' and '-'. When it's stated 'append buffer to
        list', the buffer is first converted to an int if that's possible,
        otherwise left alone as a string. It will only be appended if it's
        length is not 0.

        * If a '.' is encountered, the current buffer is appended to the current
          list, either as a int (if it's a number) or a string.
        * If a '-' is encountered, do the same as when a '.' is encountered,
          then create a new sublist, append it to the current list and replace
          the current list with the new sub-list.
        * If the last character was a digit:
            * and the current one is too, append it to the buffer.
            * otherwise append the current buffer to the list, reset the buffer
              with the current char as content
        * if the last character was NOT a digit:
            * if the current character is also NOT a digit, append it to the
              buffer
            * if it is a digit, append buffer to list, set buffers content to
              the digit
        * finally, append the buffer to the list
        """
        self._unparsed = version
        parsed = current_list = []
        buf = str(version.strip()).lower()
        start = 0
        is_digit = False
        for idx, ch in enumerate(buf):
            if ch == ".":
                if idx == start:
                    current_list.append(0)
                else:
                    current_list.append(self._parse_buffer(buf[start:idx]))
                start = idx + 1
            elif ch == "-":
                if idx == start:
                    current_list.append(0)
                else:
                    current_list.append(self._parse_buffer(buf[start:idx]))
                start = idx + 1
                current_list = self._new_list(current_list)
            elif ch.isdigit():
                if not is_digit and idx > start:
                    current_list.append(self._parse_buffer(buf[start:idx], True))
                    current_list = self._new_list(current_list)
                    start = idx
                is_digit = True
            else:
                if is_digit and idx > start:
                    current_list.append(self._parse_buffer(buf[start:idx]))
                    current_list = self._new_list(current_list)
                    start = idx
                is_digit = False
        else:
            if len(buf) > start:
                current_list.append(self._parse_buffer(buf[start:]))
        current_list = self._normalize(current_list)

        self._parsed = list2tuple(self._normalize(parsed))

    def __cmp__(self, other):
        if self is other:
            return 0

        if not isinstance(other, self.__class__):
            if isinstance(other, str):
                return self._compare(self._parsed, self.__class__(other)._parsed)
            elif isinstance(other, VersionRange) and other.version:
                return self._compare(self._parsed, other.version._parsed)
            return 1

        return self._compare(self._parsed, other._parsed)

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __hash__(self):
        return hash(self._unparsed)

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def __repr__(self):
        return "<%s.%s(%r)>" % (self.__module__, "Version", self._unparsed)

    def __str__(self):
        return self._unparsed

    def _compare(self, this, other):
        if isinstance(this, int):
            return self._int_compare(this, other)
        elif isinstance(this, str):
            return self._string_compare(this, other)
        elif isinstance(this, (list, tuple)):
            return self._list_compare(this, other)
        else:
            raise RuntimeError("Unknown type for t: %r" % this)

    def _int_compare(self, this, other):
        if isinstance(other, int):
            return this - other
        elif isinstance(other, (str, list, tuple)):
            return 1
        elif other is None:
            return this
        else:
            raise RuntimeError("other is of invalid type: %s" % type(other))

    def _list_compare(self, l, other):
        if other is None:
            if len(l) == 0:
                return 0
            return self._compare(l[0], other)
        if isinstance(other, int):
            return -1
        elif isinstance(other, str):
            return 1
        elif isinstance(other, (list, tuple)):
            for left, right in zip_longest(l, other):
                if left is None:
                    if right is None:
                        result = 0
                    else:
                        result = -1 * self._compare(right, left)
                else:
                    result = self._compare(left, right)
                if result != 0:
                    return result
            else:
                return 0
        else:
            raise RuntimeError("other is of invalid type: %s" % type(other))

    def _new_list(self, l):
        """Create a new sublist, append it to the current list and return the
        sublist

        :param list l: list to add a sublist to
        :return: the sublist
        :rtype: list
        """
        l = self._normalize(l)
        sublist = []
        l.append(sublist)
        return sublist

    def _normalize(self, l):
        for item in l[::-1]:
            if not item:
                l.pop()
            elif not isinstance(item, list):
                break
        return l

    def _string_compare(self, s, other):
        """Compare string item `s` to `other`

        :param str s: string item to compare
        :param other: other item to compare
        :type other: int, str, list or None
        """
        if other is None:
            return self._string_compare(s, "")

        if isinstance(other, (int, list, tuple)):
            return -1
        elif isinstance(other, str):
            s_value = self._string_value(s)
            other_value = self._string_value(other)
            if s_value < other_value:
                return -1
            elif s_value == other_value:
                return 0
            else:
                return 1
        else:
            raise RuntimeError("other is of invalid type: %s" % type(other))

    def _parse_buffer(self, buf, followed_by_digit=False):
        """Parse the string buf to determine if it is string or an int

        :param str buf: string to parse
        :param bool followed_by_digit: s is followed by a digit, eg. 'a1'
        :return: integer or string value of buf
        :rtype: int or str
        """
        if buf.isdigit():
            buf = int(buf)
        elif followed_by_digit and len(buf) == 1:
            if buf == "a":
                buf = "alpha"
            elif buf == "b":
                buf = "beta"
            elif buf == "m":
                buf = "milestone"

        return ALIASES.get(buf, buf)

    def _string_value(self, s):
        """Convert a string into a comparable value.

        If the string is a known qualifier, or an alias of a known qualifier,
        then return its index in the QUALIFIERS list. Otherwise return a string
        of the length of the QUALIFIERS list - s, eg. 7-foo

        :param str s: string to convert into value
        :return: value of string `s`
        :rtype: str
        """
        if s in QUALIFIERS:
            return str(QUALIFIERS.index(s) + 1)

        return "%d-%s" % (len(QUALIFIERS), s)

    @classmethod
    def fromstring(cls, spec):
        return cls(spec)



#
# Copyright (c) nexB Inc. and others.
# SPDX-License-Identifier: Apache-2.0
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

collect_ignore = ["setup.py"]



#
# Copyright (c) 2006-2019, pkgcore contributors
# SPDX-License-Identifier: BSD-3-Clause
# Version comparison utility extracted from pkgcore and further stripped down.
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

import re

from univers.utils import cmp
from univers.utils import remove_spaces

_is_gentoo_version = re.compile(
    r"^(?:\d+)(?:\.\d+)*[a-zA-Z]?(?:_(p(?:re)?|beta|alpha|rc)\d*)*$"
).match

suffix_regexp = re.compile("^(alpha|beta|rc|pre|p)(\\d*)$")

revision_regexp = re.compile(r".*(-r\d+)")

suffix_value = {"pre": -2, "p": 1, "alpha": -4, "beta": -3, "rc": -1}

"""
gentoo ebuild version comparison
"""


def is_valid(string):
    version, _ = parse_version_and_revision(remove_spaces(string))
    return _is_gentoo_version(version)


def parse_version_and_revision(version_string):
    """
    Return a tuple of (version string, revision int) given a ``version_string``.
    """
    revision = 0
    version = version_string
    match = revision_regexp.search(version_string)
    if match:
        revision = int(match.group(1)[2:])
        version = version_string[: match.span(1)[0]]

    return version, revision


def vercmp(ver1, ver2):
    """
    Compare two versions ``ver1`` and ``ver2`` and return 0, 1, or -1 according
    to the Python 2 cmp() semantics:

        Compare the two objects x and y and return an integer according to the
        outcome. The return value is negative if x < y, zero if x == y and
        strictly positive if x > y.
    """
    if not ver1:
        if not ver2:
            return 0
        else:
            return -1
    elif not ver2:
        return 1

    ver1, rev1 = parse_version_and_revision(ver1)
    ver2, rev2 = parse_version_and_revision(ver2)

    # If the versions are the same, comparing revisions will suffice.
    if ver1 == ver2:
        # revisions are equal if 0 or None (versionless cpv)
        if not rev1 and not rev2:
            return 0
        return cmp(rev1, rev2)

    # Split up the versions into dotted strings and lists of suffixes.
    parts1 = ver1.split("_")
    parts2 = ver2.split("_")

    # If the dotted strings are equal, we can skip doing a detailed comparison.
    if parts1[0] != parts2[0]:

        # First split up the dotted strings into their components.
        ver_parts1 = parts1[0].split(".")
        ver_parts2 = parts2[0].split(".")

        # Pull out any letter suffix on the final components and keep
        # them for later.
        letters = []
        for ver_parts in (ver_parts1, ver_parts2):
            if ver_parts[-1][-1].isalpha():
                letters.append(ord(ver_parts[-1][-1]))
                ver_parts[-1] = ver_parts[-1][:-1]
            else:
                # Using -1 simplifies comparisons later
                letters.append(-1)

        # OPT: Pull length calculation out of the loop
        ver_parts1_len = len(ver_parts1)
        ver_parts2_len = len(ver_parts2)

        # Iterate through the components
        for v1, v2 in zip(ver_parts1, ver_parts2):

            # If the string components are equal, the numerical
            # components will be equal too.
            if v1 == v2:
                continue

            # If one of the components begins with a "0" then they
            # are compared as floats so that 1.1 > 1.02; else ints.
            if v1[0] != "0" and v2[0] != "0":
                v1 = int(v1)
                v2 = int(v2)
            else:
                # handle the 0.060 == 0.060 case.
                v1 = v1.rstrip("0")
                v2 = v2.rstrip("0")

            # If they are not equal, the higher value wins.
            c = cmp(v1, v2)
            if c:
                return c

        if ver_parts1_len > ver_parts2_len:
            return 1
        elif ver_parts2_len > ver_parts1_len:
            return -1

        # The dotted components were equal. Let's compare any single
        # letter suffixes.
        if letters[0] != letters[1]:
            return cmp(letters[0], letters[1])

    # The dotted components were equal, so remove them from our lists
    # leaving only suffixes.
    del parts1[0]
    del parts2[0]

    # OPT: Pull length calculation out of the loop
    parts1_len = len(parts1)
    parts2_len = len(parts2)

    # Iterate through the suffixes
    for x in range(max(parts1_len, parts2_len)):

        # If we're at the end of one of our lists, we need to use
        # the next suffix from the other list to decide who wins.
        if x == parts1_len:
            match = suffix_regexp.match(parts2[x])
            val = suffix_value[match.group(1)]
            if val:
                return cmp(0, val)
            return cmp(0, int("0" + match.group(2)))
        if x == parts2_len:
            match = suffix_regexp.match(parts1[x])
            val = suffix_value[match.group(1)]
            if val:
                return cmp(val, 0)
            return cmp(int("0" + match.group(2)), 0)

        # If the string values are equal, no need to parse them.
        # Continue on to the next.
        if parts1[x] == parts2[x]:
            continue

        # Match against our regular expression to make a split between
        # "beta" and "1" in "beta1"
        match1 = suffix_regexp.match(parts1[x])
        match2 = suffix_regexp.match(parts2[x])

        # If our int'ified suffix names are different, use that as the basis
        # for comparison.
        c = cmp(suffix_value[match1.group(1)], suffix_value[match2.group(1)])
        if c:
            return c

        # Otherwise use the digit as the basis for comparison.
        c = cmp(int("0" + match1.group(2)), int("0" + match2.group(2)))
        if c:
            return c

    # Our versions had different strings but ended up being equal.
    # The revision holds the final difference.
    return cmp(rev1, rev2)



#
# Copyright (c) Chad Fowler, Rich Kilmer, Jim Weirich and others.
# Portions copyright (c) Engine Yard and Andre ArkoFacebook, Inc. and its affiliates.
#
# SPDX-License-Identifier: MIT
#
# Originally from https://github.com/rubygems/rubygems

from univers.gem import GemConstraint
from univers.gem import GemRequirement
from univers.gem import GemVersion
from univers.gem import InvalidRequirementError


def test_equals():
    refute_requirement_equal("= 1.2", "= 1.3")
    refute_requirement_equal("~> 1.3", "~> 1.3.0")
    assert_requirement_equal(["> 2", "~> 1.3", "~> 1.3.1"], ["~> 1.3.1", "~> 1.3", "> 2"])
    assert_requirement_equal(["> 2", "~> 1.3"], ["> 2.0", "~> 1.3"])


def test_initialize():
    assert_requirement_equal("= 2", "2")
    assert_requirement_equal("= 2", ["2"])
    assert_requirement_equal("= 2", GemVersion(2))
    assert_requirement_equal("2.0", "2")


def test_create():
    r = GemRequirement(">= 1", "< 2")
    assert r.constraints == (
        GemConstraint(">=", GemVersion(1)),
        GemConstraint("<", GemVersion(2)),
    )
    assert GemRequirement("= 1") == GemRequirement("= 1")
    assert GemRequirement(">= 1.2", "<= 1.3") == GemRequirement("<= 1.3", ">= 1.2")


def test_explicit_default_is_not_none():
    r = GemRequirement(">= 0")
    assert r


def test_basic_non_none():
    r = GemRequirement("= 1")
    assert r


def test_for_lockfile():
    assert GemRequirement("~> 1.0").for_lockfile() == " (~> 1.0)"
    assert GemRequirement(">= 1.0.1", "~> 1.0").for_lockfile() == " (~> 1.0, >= 1.0.1)"
    duped = GemRequirement("= 1.0", ["=", GemVersion("1.0")])
    assert duped.for_lockfile() == " (= 1.0)"


def test_parse():
    assert GemRequirement.parse("  1") == GemConstraint("=", GemVersion(1))
    assert GemRequirement.parse("= 1") == GemConstraint("=", GemVersion(1))
    assert GemRequirement.parse("> 1") == GemConstraint(">", GemVersion(1))
    assert GemRequirement.parse("=\n1") == GemConstraint("=", GemVersion(1))
    assert GemRequirement.parse("1.0") == GemConstraint("=", GemVersion(1))

    assert GemRequirement.parse(GemVersion("2")) == GemConstraint("=", GemVersion(2))


def test_parse_deduplication():
    assert GemRequirement.parse("~> 1") == GemConstraint("~>", GemVersion("1"))


def test_parse_bad():
    bads = [
        None,
        "",
        "! 1",
        "= junk",
        "1..2",
    ]
    for bad in bads:
        try:
            GemRequirement.parse(bad)
            raise Exception("exception not raised")
        except InvalidRequirementError:
            pass


def test_prerelease_eh():
    r = GemVersion("1")
    assert not r.prerelease()

    r = GemVersion("1.a")
    assert r.prerelease()

    r = GemVersion("1.x")
    assert r.prerelease()


def test_satisfied_by_eh_bang_equal():
    r = GemRequirement("!= 1.2")

    assert_satisfied_by("1.1", r)
    refute_satisfied_by("1.2", r)
    assert_satisfied_by("1.3", r)


def test_satisfied_by_eh_blank():
    r = GemRequirement("1.2")

    refute_satisfied_by("1.1", r)
    assert_satisfied_by("1.2", r)
    refute_satisfied_by("1.3", r)


def test_satisfied_by_eh_equal():
    r = GemRequirement("= 1.2")

    refute_satisfied_by("1.1", r)
    assert_satisfied_by("1.2", r)
    refute_satisfied_by("1.3", r)


def test_satisfied_by_eh_gt():
    r = GemRequirement("> 1.2")

    refute_satisfied_by("1.1", r)
    refute_satisfied_by("1.2", r)
    assert_satisfied_by("1.3", r)


def test_satisfied_by_eh_gte():
    r = GemRequirement(">= 1.2")

    refute_satisfied_by("1.1", r)
    assert_satisfied_by("1.2", r)
    assert_satisfied_by("1.3", r)


def test_satisfied_by_eh_list():
    r = GemRequirement("> 1.1", "< 1.3")

    refute_satisfied_by("1.1", r)
    assert_satisfied_by("1.2", r)
    refute_satisfied_by("1.3", r)


def test_satisfied_by_eh_lt():
    r = GemRequirement("< 1.2")

    assert_satisfied_by("1.1", r)
    refute_satisfied_by("1.2", r)
    refute_satisfied_by("1.3", r)


def test_satisfied_by_eh_lte():
    r = GemRequirement("<= 1.2")

    assert_satisfied_by("1.1", r)
    assert_satisfied_by("1.2", r)
    refute_satisfied_by("1.3", r)


def test_satisfied_by_eh_tilde_gt():
    r = GemRequirement("~> 1.2")

    refute_satisfied_by("1.1", r)
    assert_satisfied_by("1.2", r)
    assert_satisfied_by("1.3", r)


def test_satisfied_by_eh_tilde_gt_v0():
    r = GemRequirement("~> 0.0.1")

    refute_satisfied_by("0.1.1", r)
    assert_satisfied_by("0.0.2", r)
    assert_satisfied_by("0.0.1", r)


def test_satisfied_by_eh_good_problematic():
    assert_satisfied_by("0.0.1.0", "> 0.0.0.1")


def test_satisfied_by_eh_good():
    assert_satisfied_by("0.2.33", "= 0.2.33")
    assert_satisfied_by("0.2.34", "> 0.2.33")
    assert_satisfied_by("1.0", "= 1.0")
    assert_satisfied_by("1.0.0", "= 1.0")
    assert_satisfied_by("1.0", "= 1.0.0")
    assert_satisfied_by("1.0", "1.0")
    assert_satisfied_by("1.8.2", "> 1.8.0")
    assert_satisfied_by("1.112", "> 1.111")
    assert_satisfied_by("0.2", "> 0.0.0")
    assert_satisfied_by("0.0.0.0.0.2", "> 0.0.0")
    assert_satisfied_by("10.3.2", "> 9.3.2")
    assert_satisfied_by("1.0.0.0", "= 1.0")
    assert_satisfied_by("10.3.2", "!= 9.3.4")
    assert_satisfied_by("10.3.2", "> 9.3.2")
    assert_satisfied_by(" 9.3.2", ">= 9.3.2")
    assert_satisfied_by("9.3.2 ", ">= 9.3.2")
    assert_satisfied_by("", "= 0")
    assert_satisfied_by("", "< 0.1")
    assert_satisfied_by("  ", "< 0.1 ")
    assert_satisfied_by("", " <  0.1")
    assert_satisfied_by("  ", "> 0.a ")
    assert_satisfied_by("", " >  0.a")
    assert_satisfied_by("3.1", "< 3.2.rc1")

    assert_satisfied_by("3.2.0", "> 3.2.0.rc1")
    assert_satisfied_by("3.2.0.rc2", "> 3.2.0.rc1")

    assert_satisfied_by("3.0.rc2", "< 3.0")
    assert_satisfied_by("3.0.rc2", "< 3.0.0")
    assert_satisfied_by("3.0.rc2", "< 3.0.1")

    assert_satisfied_by("3.0.rc2", "> 0")

    assert_satisfied_by("5.0.0.rc2", "~> 5.a")
    refute_satisfied_by("5.0.0.rc2", "~> 5.x")

    assert_satisfied_by("5.0.0", "~> 5.a")
    assert_satisfied_by("5.0.0", "~> 5.x")


def test_illformed_requirements():
    bads = [">>> 1.3.5", "> blah"]
    for bad in bads:
        try:
            GemRequirement.parse(bad)
            raise Exception("exception not raised")
        except InvalidRequirementError:
            pass


def test_satisfied_by_eh_boxed():
    refute_satisfied_by("1.3", "~> 1.4")
    assert_satisfied_by("1.4", "~> 1.4")
    assert_satisfied_by("1.5", "~> 1.4")
    refute_satisfied_by("2.0", "~> 1.4")

    refute_satisfied_by("1.3", "~> 1.4.4")
    refute_satisfied_by("1.4", "~> 1.4.4")
    assert_satisfied_by("1.4.4", "~> 1.4.4")
    assert_satisfied_by("1.4.5", "~> 1.4.4")
    refute_satisfied_by("1.5", "~> 1.4.4")
    refute_satisfied_by("2.0", "~> 1.4.4")

    refute_satisfied_by("1.1.pre", "~> 1.0.0")
    refute_satisfied_by("1.1.pre", "~> 1.1")
    refute_satisfied_by("2.0.a", "~> 1.0")
    refute_satisfied_by("2.0.a", "~> 2.0")

    refute_satisfied_by("0.9", "~> 1")
    assert_satisfied_by("1.0", "~> 1")
    assert_satisfied_by("1.1", "~> 1")
    refute_satisfied_by("2.0", "~> 1")


def test_satisfied_by_eh_multiple():
    req = [">= 1.4", "<= 1.6", "!= 1.5"]

    refute_satisfied_by("1.3", req)
    assert_satisfied_by("1.4", req)
    refute_satisfied_by("1.5", req)
    assert_satisfied_by("1.6", req)
    refute_satisfied_by("1.7", req)
    refute_satisfied_by("2.0", req)


def test_satisfied_by_boxed():
    refute_satisfied_by("1.3", "~> 1.4")
    assert_satisfied_by("1.4", "~> 1.4")
    assert_satisfied_by("1.4.0", "~> 1.4")
    assert_satisfied_by("1.5", "~> 1.4")
    refute_satisfied_by("2.0", "~> 1.4")

    refute_satisfied_by("1.3", "~> 1.4.4")
    refute_satisfied_by("1.4", "~> 1.4.4")
    assert_satisfied_by("1.4.4", "~> 1.4.4")
    assert_satisfied_by("1.4.5", "~> 1.4.4")
    refute_satisfied_by("1.5", "~> 1.4.4")
    refute_satisfied_by("2.0", "~> 1.4.4")


def test_satisfied_by_explicitly_bounded():
    req = [">= 1.4.4", "< 1.5"]

    assert_satisfied_by("1.4.5", req)
    assert_satisfied_by("1.5.0.rc1", req)
    refute_satisfied_by("1.5.0", req)

    req = [">= 1.4.4", "< 1.5.a"]

    assert_satisfied_by("1.4.5", req)
    refute_satisfied_by("1.5.0.rc1", req)
    refute_satisfied_by("1.5.0", req)


def test_bad():
    refute_satisfied_by("", "> 0.1")
    refute_satisfied_by("1.2.3", "!= 1.2.3")
    refute_satisfied_by("1.2.003.0.0", "!= 1.02.3")
    refute_satisfied_by("4.5.6", "< 1.2.3")
    refute_satisfied_by("1.0", "> 1.1")
    refute_satisfied_by("", "= 0.1")
    refute_satisfied_by("1.1.1", "> 1.1.1")
    refute_satisfied_by("1.2", "= 1.1")
    refute_satisfied_by("1.40", "= 1.1")
    refute_satisfied_by("1.3", "= 1.40")
    refute_satisfied_by("9.3.3", "<= 9.3.2")
    refute_satisfied_by("9.3.1", ">= 9.3.2")
    refute_satisfied_by("9.3.03", "<= 9.3.2")
    refute_satisfied_by("1.0.0.1", "= 1.0")


def test_equal_with_multiple_versions():
    r1 = GemRequirement("1.0", "2.0")
    r2 = GemRequirement("2.0", "1.0")
    assert r1 == r2

    r1 = GemRequirement("1.0", "2.0", "3.0")
    r2 = GemRequirement("3.0", "1.0", "2.0")
    assert r1 == r2


def test_equivalent_requirements_are_equal():
    refute_requirement_equal("= 1.2", "= 1.3")
    refute_requirement_equal("= 1.3", "= 1.2")

    refute_requirement_equal("~> 1.3", "~> 1.3.0")
    refute_requirement_equal("~> 1.3.0", "~> 1.3")

    assert_requirement_equal(["> 2", "~> 1.3", "~> 1.3.1"], ["~> 1.3.1", "~> 1.3", "> 2"])

    assert_requirement_equal(["> 2", "~> 1.3"], ["> 2.0", "~> 1.3"])
    assert_requirement_equal(["> 2.0", "~> 1.3"], ["> 2", "~> 1.3"])

    assert_requirement_equal("= 1.0", "= 1.0.0")
    assert_requirement_equal("= 1.1", "= 1.1.0")
    assert_requirement_equal("= 1", "= 1.0.0")

    assert_requirement_equal("1.0", "1.0.0")
    assert_requirement_equal("1.1", "1.1.0")
    assert_requirement_equal("1", "1.0.0")


def assert_requirement_equal(expected, actual):
    # Assert that two requirements are equal. Handles GemRequirements,
    # strings, arrays, numbers, and versions.
    assert GemRequirement.create(actual) == GemRequirement.create(expected)


def refute_requirement_equal(unexpected, actual):
    # Refute the assumption that two requirements are equal.
    assert GemRequirement.create(actual) != GemRequirement.create(unexpected)
    assert GemRequirement.create(unexpected) != GemRequirement.create(actual)


def assert_satisfied_by(version, requirement):
    # Assert that +version+ satisfies +requirement+.
    if not isinstance(requirement, GemRequirement):
        requirement = GemRequirement.create(requirement)
    assert requirement.satisfied_by(GemVersion(version))


def refute_satisfied_by(version, requirement):
    # Refute the assumption that +version+ satisfies +requirement+.
    if not isinstance(requirement, GemRequirement):
        requirement = GemRequirement.create(requirement)
    assert not requirement.satisfied_by(GemVersion(version))



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------

project = "nexb-skeleton"
copyright = "nexB Inc. and others."
author = "AboutCode.org authors and contributors"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
'sphinx.ext.intersphinx',
]

# This points to aboutcode.readthedocs.io
# In case of "undefined label" ERRORS check docs on intersphinx to troubleshoot
# Link was created at commit - https://github.com/nexB/aboutcode/commit/faea9fcf3248f8f198844fe34d43833224ac4a83

intersphinx_mapping = {
    'aboutcode': ('https://aboutcode.readthedocs.io/en/latest/', None),
    'scancode-workbench': ('https://scancode-workbench.readthedocs.io/en/develop/', None),
}


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

master_doc = 'index'

html_context = {
    "css_files": [
        "_static/theme_overrides.css",  # override wide tables in RTD theme
    ],
    "display_github": True,
    "github_user": "nexB",
    "github_repo": "nexb-skeleton",
    "github_version": "develop",  # branch
    "conf_py_path": "/docs/source/",  # path in the checkout to the docs root
}



#
# Copyright (c) nexB Inc. and others.
# SPDX-License-Identifier: Apache-2.0
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

from unittest import TestCase
import pytest

from univers.version_constraint import VersionConstraint
from univers.version_range import GemVersionRange
from univers.version_range import InvalidVersionRange
from univers.version_range import PypiVersionRange
from univers.version_range import VersionRange
from univers.version_range import RANGE_CLASS_BY_SCHEMES
from univers.version_range import NpmVersionRange
from univers.version_range import OpensslVersionRange
from univers.versions import PypiVersion
from univers.versions import RubygemsVersion
from univers.versions import SemverVersion
from univers.versions import OpensslVersion


class TestVersionRange(TestCase):
    def test_VersionRange_afrom_string(self):
        version_range = VersionRange.from_string("vers:pypi/>0.0.2")
        assert version_range == PypiVersionRange(
            constraints=[VersionConstraint(comparator=">", version=PypiVersion(string="0.0.2"))]
        )

    def test_VersionRange_to_string(self):
        vers = "vers:pypi/0.0.2|0.0.6|>=0.0.0|0.0.1|0.0.4|0.0.5|0.0.3"
        version_range = VersionRange.from_string(vers)
        # note the sorting taking place
        assert str(version_range) == "vers:pypi/>=0.0.0|0.0.1|0.0.2|0.0.3|0.0.4|0.0.5|0.0.6"

    def test_VersionRange_pypi_does_not_contain_basic(self):
        vers = "vers:pypi/0.0.2|0.0.6|>=0.0.0|0.0.1|0.0.4|0.0.5|0.0.3"
        version_range = VersionRange.from_string(vers)
        assert not version_range.contains(PypiVersion("2.0.3"))

    def test_VersionRange_does_not_contain_version_after_range(self):
        vers = "vers:pypi/>=1.0.0|<=2.0.0"
        version_range = VersionRange.from_string(vers)
        assert not version_range.contains(PypiVersion("2.0.3"))

    def test_VersionRange_does_not_contain_version_before_range(self):
        vers = "vers:pypi/>=1.0.0|<=2.0.0"
        version_range = VersionRange.from_string(vers)
        assert not version_range.contains(PypiVersion("0.0.9"))

    def test_VersionRange_does_not_contain_version_in_between(self):
        vers = "vers:pypi/<=1.0.0|>=2.0.0"
        version_range = VersionRange.from_string(vers)
        assert not version_range.contains(PypiVersion("1.5"))

    def test_VersionRange_does_not_contain_version_excluded(self):
        vers = "vers:pypi/>=3.0.0|!=2.0.3"
        version_range = VersionRange.from_string(vers)
        assert not version_range.contains(PypiVersion("2.0.3"))

    def test_VersionRange_contains_version_after(self):
        version_range = VersionRange.from_string("vers:pypi/>0.0.2")
        assert PypiVersion("0.0.3") in version_range

    def test_VersionRange_contains_version_before(self):
        version_range = VersionRange.from_string("vers:pypi/<0.0.2")
        assert PypiVersion("0.0.0.1") in version_range

    def test_VersionRange_contains_version_included(self):
        vers = "vers:pypi/>=3.0.0|2.0.3"
        version_range = VersionRange.from_string(vers)
        assert version_range.contains(PypiVersion("2.0.3"))

    def test_VersionRange_contains_version_in_between(self):
        vers = "vers:pypi/>=1.0.0|<=2.0.0"
        version_range = VersionRange.from_string(vers)
        assert version_range.contains(PypiVersion("1.5"))

    def test_VersionRange_from_string_pypi(self):
        vers = "vers:pypi/0.0.2|0.0.6|0.0.0|0.0.1|0.0.4|0.0.5|0.0.3"
        version_range = VersionRange.from_string(vers)
        assert version_range.scheme == "pypi"
        # note the sorting taking place
        expected = (
            VersionConstraint(comparator="=", version=PypiVersion(string="0.0.0")),
            VersionConstraint(comparator="=", version=PypiVersion(string="0.0.1")),
            VersionConstraint(comparator="=", version=PypiVersion(string="0.0.2")),
            VersionConstraint(comparator="=", version=PypiVersion(string="0.0.3")),
            VersionConstraint(comparator="=", version=PypiVersion(string="0.0.4")),
            VersionConstraint(comparator="=", version=PypiVersion(string="0.0.5")),
            VersionConstraint(comparator="=", version=PypiVersion(string="0.0.6")),
        )
        assert version_range.constraints == expected
        # note the sorting taking place
        assert str(version_range) == "vers:pypi/0.0.0|0.0.1|0.0.2|0.0.3|0.0.4|0.0.5|0.0.6"

        version_range1 = VersionRange.from_string(vers, simplify=False, validate=True)
        assert version_range1.constraints == expected

        version_range2 = VersionRange.from_string(vers, simplify=True, validate=False)
        assert version_range2.constraints == expected

        version_range3 = VersionRange.from_string(vers, simplify=True, validate=True)
        assert version_range3.constraints == expected

    def test_VersionRange_from_string_pypi_complex_simplify(self):
        vers = "vers:pypi/>0.0.0|>=0.0.1|0.0.2|<0.0.3|0.0.4|<0.0.5|>=0.0.6"
        version_range = VersionRange.from_string(vers, simplify=True)
        assert str(version_range) == "vers:pypi/>0.0.0|<0.0.5|>=0.0.6"
        try:
            version_range = VersionRange.from_string(vers, validate=True)
            raise Exception(f"Exception not raised: {vers}")
        except ValueError:
            pass
        version_range = VersionRange.from_string(vers, validate=True, simplify=True)
        assert str(version_range) == "vers:pypi/>0.0.0|<0.0.5|>=0.0.6"

    def test_VersionRange_from_string_pypi_complex_simplify_and_validate(self):
        vers = "vers:pypi/>0.0.0|>=0.0.1|0.0.2|0.0.3|0.0.4|<0.0.5|>=0.0.6|!=0.8"
        version_range = VersionRange.from_string(vers, simplify=True)
        assert str(version_range) == "vers:pypi/>0.0.0|<0.0.5|>=0.0.6|!=0.8"
        version_range = VersionRange.from_string(vers, simplify=True, validate=True)

    def test_VersionRange_from_string_pypi_complex_simplify2(self):
        vers = (
            "vers:pypi/>0.0.0|>=0.0.1|>=0.0.1|0.0.2|0.0.3|0.0.4|<0.0.5|<=0.0.6|!=0.7|8.0|>12|<15.3"
        )
        version_range = VersionRange.from_string(vers, simplify=True)
        assert str(version_range) == "vers:pypi/>0.0.0|<=0.0.6|!=0.7|8.0|>12|<15.3"

    def test_VersionRange_from_string_pypi_simple_cases(self):
        vers = "vers:pypi/>0.0.1"
        version_range = VersionRange.from_string(vers, simplify=True, validate=True)
        assert str(version_range) == vers

        vers = "vers:pypi/>=0.0.1"
        version_range = VersionRange.from_string(vers, simplify=True, validate=True)
        assert str(version_range) == vers

        vers = "vers:pypi/<0.0.1"
        version_range = VersionRange.from_string(vers, simplify=True, validate=True)
        assert str(version_range) == vers

        vers = "vers:pypi/<=0.0.1"
        version_range = VersionRange.from_string(vers, simplify=True, validate=True)
        assert str(version_range) == vers

        vers = "vers:pypi/0.0.1"
        version_range = VersionRange.from_string(vers, simplify=True, validate=True)
        assert str(version_range) == vers

        vers = "vers:pypi/!=0.0.1"
        version_range = VersionRange.from_string(vers, simplify=True, validate=True)
        assert str(version_range) == vers

        vers = "vers:pypi/*"
        version_range = VersionRange.from_string(vers, simplify=True, validate=True)
        assert str(version_range) == vers

    def test_VersionRange_from_string_pypi_two_cases(self):
        vers = "vers:pypi/>0.0.1|<0.1"
        version_range = VersionRange.from_string(vers, simplify=True, validate=True)
        assert str(version_range) == vers

        vers = "vers:pypi/>=0.0.1|<0.1"
        version_range = VersionRange.from_string(vers, simplify=True, validate=True)
        assert str(version_range) == vers

        vers = "vers:pypi/<0.0.1|>0.1"
        version_range = VersionRange.from_string(vers, simplify=True, validate=True)
        assert str(version_range) == vers

        vers = "vers:pypi/<=0.0.1|>0.1"
        version_range = VersionRange.from_string(vers, simplify=True, validate=True)
        assert str(version_range) == vers

        vers = "vers:pypi/0.0.1|>0.1"
        version_range = VersionRange.from_string(vers, simplify=True, validate=True)
        assert str(version_range) == vers

        vers = "vers:pypi/!=0.0.1|>0.1"
        version_range = VersionRange.from_string(vers, simplify=True, validate=True)
        assert str(version_range) == vers

    def test_GemVersionRange_from_native_range_with_pessimistic_operator(self):
        gem_range = "~>2.0.8"
        version_range = GemVersionRange.from_native(gem_range)
        assert version_range.to_string() == "vers:gem/>=2.0.8|<2.1"
        assert version_range.constraints == (
            VersionConstraint(comparator=">=", version=RubygemsVersion(string="2.0.8")),
            VersionConstraint(comparator="<", version=RubygemsVersion(string="2.1")),
        )

    def test_VersionRange_contains_works_for_star_range(self):
        from univers.versions import SemverVersion

        assert SemverVersion("1.0.0") in VersionRange.from_string("vers:nginx/*")

    def test_PypiVersionRange_raises_ivr_for_unsupported_ranges(self):
        try:
            PypiVersionRange.from_native(
                "~= 0.9, >= 1.0, != 1.3.4.*, < 2.0, ~= 1.3.4.*, ===1.0, ==1.*"
            )
            raise Exception("Exception not raised")
        except InvalidVersionRange as ivre:
            assert str(ivre).startswith("Unsupported character")

    def test_PypiVersionRange_raises_ivr_for_invalid_ranges(self):
        try:
            PypiVersionRange.from_native("~= 1.3, ===1.0, ==1.*")
            raise Exception("Exception not raised")
        except InvalidVersionRange as ivre:
            assert str(ivre).startswith("Unsupported character")

    def test_NpmVersionRange_from_native_with_compatible_with_version_operator(self):
        npm_range = "^1.2.9"
        expected = NpmVersionRange(
            constraints=(
                VersionConstraint(comparator=">=", version=SemverVersion(string="1.2.9")),
                VersionConstraint(comparator="<", version=SemverVersion(string="2.0.0")),
            )
        )
        version_range = NpmVersionRange.from_native(npm_range)
        assert version_range == expected

    def test_NpmVersionRange_from_native_with_approximately_equal_to_operator(self):
        npm_range = "~3.8.2"
        expected = NpmVersionRange(
            constraints=(
                VersionConstraint(comparator=">=", version=SemverVersion(string="3.8.2")),
                VersionConstraint(comparator="<", version=SemverVersion(string="3.9.0")),
            )
        )
        version_range = NpmVersionRange.from_native(npm_range)
        assert version_range == expected

    def test_OpensslVersionRange_from_native_single_legacy(self):
        openssl_range = "0.9.8j"
        expected = OpensslVersionRange(
            constraints=(
                VersionConstraint(comparator="=", version=OpensslVersion(string="0.9.8j")),
            )
        )
        version_range = OpensslVersionRange.from_native(openssl_range)
        assert version_range == expected

    def test_OpensslVersionRange_from_native_single_new_semver(self):
        openssl_range = "3.0.1"
        expected = OpensslVersionRange(
            constraints=(VersionConstraint(comparator="=", version=OpensslVersion(string="3.0.1")),)
        )
        version_range = OpensslVersionRange.from_native(openssl_range)
        assert version_range == expected

    def test_OpensslVersionRange_from_native_mixed(self):
        openssl_range = "3.0.0, 1.0.1b"
        expected = OpensslVersionRange(
            constraints=(
                VersionConstraint(comparator="=", version=OpensslVersion(string="1.0.1b")),
                VersionConstraint(comparator="=", version=OpensslVersion(string="3.0.0")),
            )
        )
        version_range = OpensslVersionRange.from_native(openssl_range)
        assert version_range == expected

    def test_OpensslVersionRange_from_versions(self):
        sequence = ["3.0.0", "1.0.1b", "3.0.2", "0.9.7a ", "1.1.1ka"]
        expected = OpensslVersionRange(
            constraints=(
                VersionConstraint(comparator="=", version=OpensslVersion(string="0.9.7a")),
                VersionConstraint(comparator="=", version=OpensslVersion(string="1.0.1b")),
                VersionConstraint(comparator="=", version=OpensslVersion(string="1.1.1ka")),
                VersionConstraint(comparator="=", version=OpensslVersion(string="3.0.0")),
                VersionConstraint(comparator="=", version=OpensslVersion(string="3.0.2")),
            )
        )
        version_range = OpensslVersionRange.from_versions(sequence)
        assert version_range == expected


VERSION_RANGE_TESTS_BY_SCHEME = {
    "nginx": ["0.8.40+", "0.7.52-0.8.39", "0.9.10", "1.5.0+, 1.4.1+"],
    "npm": ["^1.2.9", "~3.8.2", "5.0.0 - 7.2.3", "2.1 || 2.6", "1.1.2 1.2.2", "<=2.1 >=1.1"],
    "openssl": ["1.1.1ak", "1.1.0", "3.0.2", "3.0.1, 0.9.7a", "1.0.2ck, 3.1.2"],
}


@pytest.mark.xfail(reason="Not all schemes are implemented yet")
def test_all_schemes_are_tested_for_round_tripping():
    untested_schemes = []
    for scheme in RANGE_CLASS_BY_SCHEMES:
        if scheme not in VERSION_RANGE_TESTS_BY_SCHEME:
            untested_schemes.append(scheme)
    assert not untested_schemes


@pytest.mark.parametrize(
    "scheme, native_ranges",
    VERSION_RANGE_TESTS_BY_SCHEME.items(),
)
def test_from_native_and_from_string_round_trip(scheme, native_ranges):

    range_class = RANGE_CLASS_BY_SCHEMES[scheme]
    for rng in native_ranges:
        from_native = range_class.from_native(rng)
        from_string = range_class.from_string(from_native.to_string())
        assert from_native == from_string



#
# Copyright (c) nexB Inc. and others.
# SPDX-License-Identifier: Apache-2.0
#
# Visit https://aboutcode.org and https://github.com/nexB/univers for support and download.

import attr
import semantic_version
from packaging.specifiers import InvalidSpecifier
from packaging.specifiers import SpecifierSet
from semantic_version.base import AllOf
from semantic_version.base import AnyOf

from univers import gem
from univers import versions
from univers.utils import remove_spaces
from univers.version_constraint import VersionConstraint
from univers.version_constraint import contains_version


class InvalidVersionRange(Exception):
    """
    Error for scheme-specific version syntax is not supported or not valid
    """


@attr.s(frozen=True, order=False, eq=True, hash=True)
class VersionRange:
    """
    Base version range class. Subclasses must provide implement.
    A VersionRange represents a list of constraints on the versions "timeline"
    of a package.
    """

    # Versioning scheme. By convention this is the same as the Package URL
    # package type since this "defines" most commonly the versioning scheme of a
    # package, such as "npm" (whose scheme is defined in the "node-semver" npm
    # package. This could be something else though and a purl may be accompanied
    # by a version range for another scheme; for example, a
    # ``pkg:github/foo/bar`` purl could be accompanied by a a ``vers:npm/12.3``
    # range. Subclasses MUST provide this.
    scheme = None

    # Version subclass to use with this versioning scheme, such as
    # PypiVersion. Subclasses MUST provide this.
    version_class = None

    # A tuple of VersionConstraint that are signposts on the versions
    # timeline
    constraints = attr.ib(type=tuple, default=attr.Factory(tuple))

    def __attrs_post_init__(self, *args, **kwargs):
        constraints = tuple(sorted(self.constraints))
        # Notes: setattr is used because this is an immutable frozen instance.
        # See https://www.attrs.org/en/stable/init.html?#post-init
        object.__setattr__(self, "constraints", constraints)

    @classmethod
    def from_native(cls, string):
        """
        Return a VersionRange built from a scheme-specific, native version range
        ``string``. Subclasses can implement.
        """
        return NotImplementedError

    @classmethod
    def from_natives(cls, strings):
        """
        Return a VersionRange built from a ``strings`` list of scheme-
        specific native version range strings. Subclasses can implement.
        """
        return NotImplementedError

    def to_native(self, *args, **kwargs):
        """
        Return a native range string for this VersionRange. Subclasses can
        implement. Optional ``args`` and ``kwargs`` allow subclass to require
        extra arguments (such as a package name that some scheme may require
        like for deb and rpm.)
        """
        return NotImplementedError

    @classmethod
    def from_string(cls, vers, simplify=False, validate=False):
        """
        Return a VersionRange built from a ``vers`` version range spec string,
        such as "vers:npm/1.2.3,>=2.0.0"
        """
        # Spaces are not significant and removed in a canonical form.
        vers = remove_spaces(vers)

        # A version range specifier contains only printable ASCII letters, digits and
        # punctuation.
        is_ascii = len(vers) + 2 == len(ascii(vers))
        if not is_ascii:
            raise ValueError(f"Invalid non ASCII characters: {vers!r}")

        # The URI scheme and versioning scheme are always lowercase as in  ``vers:npm``.
        uri_scheme, _, scheme_range_spec = vers.partition(":")
        uri_scheme = uri_scheme.lower()

        if uri_scheme != "vers":
            raise ValueError(f"{vers!r} must start with the 'vers:' URI scheme.")

        versioning_scheme, _, constraints = scheme_range_spec.partition("/")
        versioning_scheme = versioning_scheme.lower()
        range_class = RANGE_CLASS_BY_SCHEMES.get(versioning_scheme)
        if not range_class:
            raise ValueError(
                f"{vers!r} has an unknown versioning scheme: " f"{versioning_scheme!r}.",
            )

        version_class = range_class.version_class

        constraints = remove_spaces(constraints)
        if not constraints:
            raise ValueError(f"{vers!r} specifies no version range constraints.")

        # There is only one star: "*" must only occur once and alone in a range,
        # without any other constraint or version.
        if constraints.startswith("*"):
            if constraints != "*":
                raise ValueError(f"{vers!r} contains an invalid '*' constraint.")
            return range_class(
                [VersionConstraint.from_string(string="*", version_class=version_class)]
            )

        parsed_constraints = []

        constraints = constraints.strip("|")
        for const in constraints.split("|"):
            constraint = VersionConstraint.from_string(
                string=const,
                version_class=version_class,
            )
            parsed_constraints.append(constraint)

        # Constraints are sorted by version**. The canonical ordering is the versions
        # order. The ordering of ``<version-constraint>`` is not significant otherwise
        # but this sort order is needed when check if a version is contained in a range.
        parsed_constraints.sort()

        if simplify:
            parsed_constraints = VersionConstraint.simplify(parsed_constraints)
        if validate:
            VersionConstraint.validate(parsed_constraints)

        return range_class(parsed_constraints)

    @classmethod
    def from_versions(cls, sequence):
        """
        Return a VersionRange built from a list of version strings,
        such as ["3.0.0", "1.0.1b", "3.0.2", "0.9.7a", "1.1.1ka"]
        """
        if not cls.scheme or not cls.version_class:
            return NotImplementedError

        constraints = []
        for version in sequence:
            version_obj = cls.version_class(version)
            constraint = VersionConstraint(comparator="=", version=version_obj)
            constraints.append(constraint)
        return cls(constraints=constraints)

    def __str__(self):
        constraints = "|".join(str(c) for c in sorted(self.constraints))
        return f"vers:{self.scheme}/{constraints}"

    to_string = __str__

    def to_dict(self):
        constraints = [c.to_dict() for c in self.constraints]
        return dict(scheme=self.scheme, constraints=constraints)

    def __contains__(self, version):
        """
        Return True if this VersionRange contains the ``version`` Version
        object. A version is contained in a VersionRange if it satisfies its
        constraints according to ``vers`` rules.
        """
        if not isinstance(version, self.version_class):
            raise TypeError(
                f"{version!r} is not of expected type: {self.version_class!r}",
            )
        return contains_version(version, self.constraints)

    contains = __contains__

    def __eq__(self, other):
        return (
            self.scheme == other.scheme
            and self.version_class == other.version_class
            and self.constraints == other.constraints
        )


def from_cve_v4(data, scheme):
    """
    Return a VersionRange build from the provided CVE V4 API ``data`` using the
    provided versioning vers ``scheme``.
    """


def from_cve_v5(data, scheme):
    """
    Return a VersionRange build from the provided CVE V5 API ``data`` using the
    provided versioning vers ``scheme``.

    See https://github.com/CVEProject/cve-schema/tree/master/schema/v5.0
    ``data`` can be:
    - a mapping of collectionURL and versions:
        {"collectionURL": "some URL", "versions": [{"versionValue": "1.0"}]}

    """


def from_osv_v1(data, scheme):
    """
    Return a VersionRange build from the provided CVE V4 API data using the
    provided versioning vers ``scheme``.
    """


class NpmVersionRange(VersionRange):
    scheme = "npm"
    version_class = versions.SemverVersion

    vers_by_native_comparators = {
        "==": "=",
        "<=": "<=",
        ">=": ">=",
        "<": "<",
        ">": ">",
    }

    @classmethod
    def from_native(cls, string):
        """
        Return a VersionRange built from an npm "node-semver" range ``string``.
        """
        # FIXME: code is entirely duplicated with the GemVersionRange

        # an NpmSpec handles parsing of both the semver versions and node-semver
        # ranges at once
        spec = semantic_version.NpmSpec(string)

        clause = spec.clause.simplify()
        assert isinstance(clause, (AnyOf, AllOf))
        anyof_constraints = []
        if isinstance(clause, AnyOf):
            for allof_clause in clause.clauses:
                anyof_constraints.extend(get_allof_constraints(cls, allof_clause))
        elif isinstance(clause, AllOf):
            alloc = get_allof_constraints(cls, clause)
            anyof_constraints.extend(alloc)
        else:
            raise ValueError(f"Unknown clause type: {spec!r}")

        return cls(constraints=anyof_constraints)


def get_allof_constraints(cls, clause):
    """
    Return a list of VersionConstraint given an AllOf ``clause``.
    """
    assert isinstance(clause, AllOf)
    allof_constraints = []
    for constraint in clause.clauses:
        comparator = cls.vers_by_native_comparators[constraint.operator]
        version = cls.version_class(str(constraint.target))
        constraint = VersionConstraint(comparator=comparator, version=version)
        allof_constraints.append(constraint)
    return allof_constraints


class GemVersionRange(VersionRange):
    """
    A version range implementation for Rubygems.

    gem need its own versioning scheme as this is not semver.
    See https//github.com/nexB/univers/issues/5
    See https://github.com/ruby/ruby/blob/415671a28273e5bfbe9aa00a0e386f025720ac23/lib/rubygems/requirement.rb
    See https//semver.org/spec/v2.0.0.html#spec-item-11
    See https//snyk.io/blog/differences-in-version-handling-gems-and-npm/
    See https://github.com/npm/node-semver/issues/112
    """

    scheme = "gem"
    version_class = versions.RubygemsVersion

    vers_by_native_comparators = {
        "=": "=",
        "!=": "!=",
        "<=": "<=",
        ">=": ">=",
        "<": "<",
        ">": ">",
    }

    @classmethod
    def from_native(cls, string):
        """
        Return a VersionRange built from a Rubygem version range ``string``.

        Gem version semantics are different from semver: there can be commonly
        more than three segments and the operators are also different.
        """

        gr = gem.GemRequirement.from_string(string).simplify()

        constraints = []
        for gc in gr.constraints:
            version = cls.version_class(str(gc.version))
            op = cls.vers_by_native_comparators[gc.op]
            vc = VersionConstraint(comparator=op, version=version)
            constraints.append(vc)

        return cls(constraints=constraints)


def split_req(string, comparators, default=None, strip=""):
    """
    Return a tuple of (vers comparator, version) strings given an common version
    requirement``string`` such as "> 2.3" or "<= 2.3" using the ``comparators``
    mapping of {native comparator: vers comparator}. Strip the ``string`` from
    the provided leading of training characters in ``strip``.

    If there is none of the ``comparators`` found in ``string``:

    - Return the ``default`` vers comparator string if provided.
    - Otherwise, raise a ValueError for an unknown comparator.

    For example::

    >>> comps = {"=": "=", "<=": "<=", ">=": ">="}
    >>> assert split_req("= 2.3", comparators=comps) == ("=", "2.3",)
    >>> assert split_req("  <   =  2 . 3  ", comparators=comps) == ("<=", "2.3",)
    >>> assert split_req(">= 2.3", comparators=comps) == (">=", "2.3",)
    >>> assert split_req(">= 2.3", comparators=comps) == (">=", "2.3",)
    >>> assert split_req("<= 2.3", comparators=comps) == ("<=", "2.3",)
    >>> assert split_req("(< =  2.3 )", comparators=comps, strip=")(") == ("<=", "2.3",)

    With a default, we return the default comparator::

    >>> assert split_req("2.3,", comparators=comps, default="=", strip=",") == ("=", "2.3",)

    Otherwise, a ValuaeError::

    >>> try:
    ...     split_req("~2.3", comparators=comps, )
    ...     raise Exception("ValueError should be raised")
    ... except ValueError:
    ...     pass
    """
    constraint_string = remove_spaces(string).strip(strip)

    for native_comparator, vers_comparator in comparators.items():
        if constraint_string.startswith(native_comparator):
            version = constraint_string.lstrip(native_comparator)
            return vers_comparator, version

    if default:
        return default, constraint_string

    raise ValueError(f"Unknown comparator in version requirement: {string!r} ")


class DebianVersionRange(VersionRange):
    """
    Debian version ranges as seen in Debian manual for relationships:
    https://www.debian.org/doc/debian-policy/ch-relationships.html

    These are for defined one expression at a time. Multiple expressions each
    com with a package name. Therefore there is no "range string" per se, instead
    there is always a list of version constraints as an input. For instance::

        libc6 (>> 2.23), libc6 (<< 2.24)'

    Therefore native conversions are different.
    """

    scheme = "deb"
    version_class = versions.DebianVersion
    vers_by_native_comparators = {
        "=": "=",
        "<=": "<=",
        ">=": ">=",
        "<<": "<",
        ">>": ">",
        # legacy
        "<": "<",
        ">": ">",
    }

    @classmethod
    def split(cls, string):
        """
        Return a tuple of (vers comparator, version) strings given a Debian
        version relationship ``string`` such as ">>2.3" or "(<< 2.3)". Raise a
        ValueError for unknown comparators.

        For example::
        >>> assert DebianVersionRange.split("=2.3") == ("=", "2.3",)
        >>> assert DebianVersionRange.split("  <   =  2 . 3  ") == ("<=", "2.3",)
        >>> assert DebianVersionRange.split("(>=2.3)") == (">=", "2.3",)
        >>> assert DebianVersionRange.split(">=2.3") == (">=", "2.3",)
        >>> assert DebianVersionRange.split("<=2.3") == ("<=", "2.3",)
        >>> assert DebianVersionRange.split("<<2.3") == ("<", "2.3",)
        >>> assert DebianVersionRange.split(">>2.3") == (">", "2.3",)
        >>> assert DebianVersionRange.split(">2.3") == (">", "2.3",)
        >>> assert DebianVersionRange.split("<2.3") == ("<", "2.3",)
        >>> try:
        ...     DebianVersionRange.split("~2.3")
        ...     raise Exception("ValueError should be raised")
        ... except ValueError:
        ...     pass
        """
        return split_req(
            string=string,
            comparators=cls.vers_by_native_comparators,
            strip=")(",
        )

    @classmethod
    def build_constraint_from_string(cls, string):
        """
        Return a VersionConstraint built from a single Debian version
        relationship ``string``.

        >>> vr = DebianVersionRange.build_constraint_from_string("= 5.0")
        >>> assert str(vr) == "5.0"
        >>> vr = DebianVersionRange.build_constraint_from_string("(>> 2.23)")
        >>> assert str(vr) == ">2.23"
        >>> vr = DebianVersionRange.build_constraint_from_string("<= 2.24")
        >>> assert str(vr) == "<=2.24"
        """
        comparator, version = cls.split(string)
        version = cls.version_class(version)
        return VersionConstraint(comparator=comparator, version=version)

    @classmethod
    def from_native(cls, string):
        """
        Return a VersionRange built from a ``string`` single Debian
        version relationship string.

        For example::

        >>> vr = DebianVersionRange.from_native("(= 3.5.6)")
        >>> assert str(vr) == "vers:deb/3.5.6"
        """
        return cls(constraints=[cls.build_constraint_from_string(string)])

    @classmethod
    def from_natives(cls, strings):
        """
        Return a VersionRange built from a ``strings`` list of Debian
        version relationships or a single relationship string.

        For example::

        >>> vr = DebianVersionRange.from_natives("= 3.5.6")
        >>> assert str(vr) == "vers:deb/3.5.6"

        >>> rels = ["(>= 2.8.16)"]
        >>> vr = DebianVersionRange.from_natives(rels)
        >>> assert str(vr) == "vers:deb/>=2.8.16"

        >>> rels = [">= 1:1.1.4", "(>= 2.8.16)", "<= 2.8.16-z"]
        >>> vr = DebianVersionRange.from_natives(rels)
        >>> assert str(vr) == "vers:deb/>=2.8.16|<=2.8.16-z|>=1:1.1.4"

        >>> rels = ["(>= 2:4.13.1)", "(<= 2:4.13.1-0ubuntu0.16.04.1.1~)"]
        >>> vr = DebianVersionRange.from_natives(rels)
        >>> assert str(vr) == "vers:deb/>=2:4.13.1|<=2:4.13.1-0ubuntu0.16.04.1.1~"

        >>> rels = ["= 5.0", "(>> 2.23)", "< 2.24"]
        >>> vr = DebianVersionRange.from_natives(rels)
        >>> assert str(vr) == "vers:deb/>2.23|<2.24|5.0"

        >>> rels = ["(<< 3:1.1.25~)", "(>> 2:1.1.24~)"]
        >>> vr = DebianVersionRange.from_natives(rels)
        >>> assert str(vr) == "vers:deb/>2:1.1.24~|<3:1.1.25~"
        """

        if isinstance(strings, str):
            return cls.from_native(strings)
        constraints = [cls.build_constraint_from_string(rel) for rel in strings]
        return cls(constraints=constraints)


class PypiVersionRange(VersionRange):
    """
    PyPI PEP 440 version range.

    For example:
    >>> from univers.versions import PypiVersion
    >>> constraints = [
    ...    VersionConstraint(version=PypiVersion("2")),
    ...    VersionConstraint(comparator=">=", version=PypiVersion("3")),
    ...    VersionConstraint(comparator="<", version=PypiVersion("4")),
    ...    VersionConstraint(version=PypiVersion("5")),
    ... ]
    >>> range = PypiVersionRange(constraints=constraints)
    >>> assert str(range) == "vers:pypi/2|>=3|<4|5"
    """

    scheme = "pypi"
    version_class = versions.PypiVersion

    vers_by_native_comparators = {
        # 01.01.01 is equal 1.1.1 e.g., with version normalization
        "==": "=",
        "!=": "!=",
        "<=": "<=",
        ">=": ">=",
        "<": "<",
        ">": ">",
        # per https://www.python.org/dev/peps/pep-0440/#compatible-release
        # For a given release identifier V.N, the compatible release clause is
        # approximately equivalent to the pair of comparison clauses:
        # >= V.N, == V.*
        "~=": None,
        # 01.01.01 is NOT equal to 1.1.1 using === which is strict string
        # equality this is a rare and eventually non-suggested approach
        "===": None,
    }

    @classmethod
    def from_native(cls, string):
        """
        Return a VersionRange built from a PyPI PEP440 version specifiers ``string``.
        Raise an a univers.versions.InvalidVersion
        """
        # TODO: environment markers are yet supported
        # TODO: handle  .* version, ~= and === operators

        if ";" in string:
            raise InvalidVersionRange(f"Unsupported PyPI environment marker: {string!r}")

        unsupported_chars = ";<>!=\\/|{}()`?'\"\t\n "
        if any(c in string for c in unsupported_chars):
            raise InvalidVersionRange(
                f"Unsupported character: {unsupported_chars!r} " f"in PyPI version: {string!r}"
            )

        try:
            specifiers = SpecifierSet(string)
        except InvalidSpecifier as e:
            raise InvalidVersionRange() from e

        # Note that in PyPI all constraints apply

        constraints = []
        unsupported_messages = []
        for spec in specifiers:
            operator = spec.operator
            version = spec.version

            if operator == "~=" or operator == "===":
                msg = f"Unsupported PyPI version constraint operator: {spec!r}"
                unsupported_messages.append(msg)

            if str(version).endswith(".*"):
                msg = f"Unsupported PyPI version: {spec!r}"
                unsupported_messages.append(msg)

            try:
                version = cls.version_class(version)
                comparator = cls.vers_by_native_comparators[operator]
                constraint = VersionConstraint(comparator=comparator, version=version)
                constraints.append(constraint)
            except:
                msg = f"Invalid PyPI version: {spec!r}"
                unsupported_messages.append(msg)

        if unsupported_messages:
            raise InvalidVersionRange(*unsupported_messages)

        return cls(constraints=constraints)


class MavenVersionRange(VersionRange):
    """
    Maven version range as documented at
    https://maven.apache.org/enforcer/enforcer-rules/versionRanges.html
    """

    scheme = "maven"
    version_class = versions.MavenVersion


class NugetVersionRange(VersionRange):
    """
    NuGet range as in:[3.10.1,4)
    """

    scheme = "nuget"
    version_class = versions.NugetVersion


class ComposerVersionRange(VersionRange):
    # TODO composer may need its own scheme see https//github.com/nexB/univers/issues/5
    # and https//getcomposer.org/doc/articles/versions.md
    scheme = "composer"
    version_class = versions.SemverVersion


class RpmVersionRange(VersionRange):
    # http://ftp.rpm.org/api/4.4.2.2/dependencies.html
    # http://ftp.rpm.org/max-rpm/s1-rpm-depend-manual-dependencies.html
    scheme = "rpm"
    version_class = versions.RpmVersion

    vers_by_native_comparators = {
        "=": "=",
        "<=": "<=",
        ">=": ">=",
        "<": "<",
        ">": ">",
        # seen in RPM code but never seen in the doc or in the wild so far
        "<>": "!=",
        # seen in a specfile parser code
        "!=": "!=",
        "==": "=",
    }

    @classmethod
    def build_constraint_from_string(cls, string):
        """
        Return a VersionConstraint built from a single RPM version
        relationship ``string``.

        >>> vr = RpmVersionRange.build_constraint_from_string("= 5.0")
        >>> assert str(vr) == "5.0", str(vr)
        >>> vr = RpmVersionRange.build_constraint_from_string("> 2.23,")
        >>> assert str(vr) == ">2.23", str(vr)
        >>> vr = RpmVersionRange.build_constraint_from_string("<= 2.24")
        >>> assert str(vr) == "<=2.24", str(vr)
        """
        comparator, version = split_req(
            string=string,
            comparators=cls.vers_by_native_comparators,
            strip=",",
        )
        version = cls.version_class(version)
        return VersionConstraint(comparator=comparator, version=version)

    @classmethod
    def from_native(cls, string):
        """
        Return a VersionRange built from a ``string`` single RPM
        version requirement string.

        For example::

        >>> vr = RpmVersionRange.from_native("= 3.5.6")
        >>> assert str(vr) == "vers:rpm/3.5.6", str(vr)
        """
        return cls(constraints=[cls.build_constraint_from_string(string)])

    @classmethod
    def from_natives(cls, strings):
        """
        Return a VersionRange built from a ``strings`` list of RPM
        version requirements or a single requirement string.

        For example::

        >>> vr = RpmVersionRange.from_natives("= 3.5.6")
        >>> assert str(vr) == "vers:rpm/3.5.6", str(vr)

        >>> reqs = [">= 2.8.16"]
        >>> vr = RpmVersionRange.from_natives(reqs)
        >>> assert str(vr) == "vers:rpm/>=2.8.16", str(vr)

        >>> reqs = [">= 1:1.1.4", ">= 2.8.16", "<= 2.8.16-z"]
        >>> vr = RpmVersionRange.from_natives(reqs)
        >>> assert str(vr) == "vers:rpm/>=2.8.16|<=2.8.16-z|>=1:1.1.4", str(vr)

        >>> reqs = ["= 5.0", "> 2.23,", "< 2.24"]
        >>> vr = RpmVersionRange.from_natives(reqs)
        >>> assert str(vr) == "vers:rpm/>2.23|<2.24|5.0", str(vr)
        """

        if isinstance(strings, str):
            return cls.from_native(strings)
        constraints = [cls.build_constraint_from_string(rel) for rel in strings]
        return cls(constraints=constraints)


class GolangVersionRange(VersionRange):
    """
    Go modules use strict semver with pseudo numbering for Git repos
    https://go.dev/doc/modules/version-numbers
    """

    scheme = "golang"
    version_class = versions.SemverVersion


class GenericVersionRange(VersionRange):
    scheme = "generic"
    version_class = versions.SemverVersion


class ApacheVersionRange(VersionRange):
    # apache is not semver at large. And in particular we may have schemes that
    # are package name-specific
    scheme = "apache"
    version_class = versions.SemverVersion


class HexVersionRange(VersionRange):
    scheme = "hex"
    version_class = versions.SemverVersion


class CargoVersionRange(VersionRange):
    scheme = "cargo"
    version_class = versions.SemverVersion


class MozillaVersionRange(VersionRange):
    scheme = "mozilla"
    version_class = versions.SemverVersion


class GitHubVersionRange(VersionRange):
    scheme = "github"
    version_class = versions.SemverVersion


class EbuildVersionRange(VersionRange):
    scheme = "ebuild"
    version_class = versions.GentooVersion


class AlpineLinuxVersionRange(VersionRange):
    scheme = "alpine"
    version_class = versions.AlpineLinuxVersion


class ArchLinuxVersionRange(VersionRange):
    scheme = "archlinux"
    version_class = versions.ArchLinuxVersion


class NginxVersionRange(VersionRange):
    """
    Nginx versioning is semver for version and their own syntax for ranges as
    used in their security advisories.

    The documentation on these ranges is minimal. See these for details:
    - https://mailman.nginx.org/pipermail/nginx/2021-September/061039.html
    - https://nginx.org/en/security_advisories.html
    - https://serverfault.com/questions/715049/what-s-the-difference-between-the-mainline-and-stable-branches-of-nginx

    In particular for versions:
    - the versions are semver.

    - versions can be in the one "mainline" branch or one of many "stable" branches.

    - for versions in the "mainline" branch, (e.g., development) the minor
      segment is an odd number.

    - versions in the "stable" branch, (e.g., a release branch) the minor
      segment is an even number. Installation are typically made from branch and
      its versions.

      For example: in 0.6.18 the 6 e.g., semver "minor" segment is either odd or even
      - odd (as with "7") means this is the "mainline" branch
      - even (as with "4") means this is in a "stable" branch

    And for ranges, we have these notations:

    - dash ranges: 0.6.18-1.20.0 where start and end are included in the range

    - comma ranges: 1.21.0+, 1.20.1+ where any of the condition applies

    - plus suffixes: 1.21.0+ where this or any later version in the branch applies
      Therefore:
      - 1.21.0+ would expand to >=1.21.0 because 21 is odd and this is the
        mainline branch

      - 1.22.0+ would expand to >=1.22.0,<1.23.0 because 22 is even and this is
        one of the stable branches

    There are two special version range values:
      - "all" means all versions.
      - "none" means no version and therefore no version range. It is used only
        in one advisory for CVE-2009-4487 and triggers an error.

    Some vulnerable ranges are only for Windows builds but the range syntax is
    the same. This could be resolved with a specific purl qualifier.
    These are prefixed by the string "nginx/Window".
    """

    scheme = "nginx"
    version_class = versions.SemverVersion

    vers_by_native_comparators = {
        "==": "=",
        "<=": "<=",
        ">=": ">=",
        "<": "<",
        ">": ">",
    }

    @classmethod
    def from_native(cls, string):
        """
        Return a VersionRange built from an nginx range ``string``.

        For example:
        >>> result = NginxVersionRange.from_native("1.5.10")
        >>> assert str(result) == "vers:nginx/1.5.10", str(result)

        >>> result = NginxVersionRange.from_native("0.7.52-0.8.39")
        >>> assert str(result) == "vers:nginx/>=0.7.52|<=0.8.39", str(result)

        >>> result = NginxVersionRange.from_native("1.1.4-1.2.8, 1.3.9-1.4.0")
        >>> assert str(result) == "vers:nginx/>=1.1.4|<=1.2.8|>=1.3.9|<=1.4.0", str(result)

        >>> result = NginxVersionRange.from_native("0.8.40+, 0.7.66+")
        >>> assert str(result) == "vers:nginx/>=0.7.66|>=0.8.40|<0.9.0", str(result)

        >>> result = NginxVersionRange.from_native("1.5.0+, 1.4.1+")
        >>> assert str(result) == "vers:nginx/>=1.4.1|<1.5.0|>=1.5.0", str(result)

        >>> result = NginxVersionRange.from_native("all")
        >>> assert str(result) == "vers:nginx/*", str(result)

        >>> try:
        ...     NginxVersionRange.from_native("none")
        ... except ValueError:
        ...     pass
        """
        cleaned = remove_spaces(string).lower()
        if cleaned == "all":
            return cls(
                constraints=[VersionConstraint(comparator="*", version_class=cls.version_class)]
            )

        constraints = []

        for clauses in cleaned.split(","):

            if "-" in clauses:
                # dash range
                start, _, end = clauses.partition("-")
                start_version = cls.version_class(start)
                end_version = cls.version_class(end)
                vstart = VersionConstraint(comparator=">=", version=start_version)
                vend = VersionConstraint(comparator="<=", version=end_version)
                constraints.extend([vstart, vend])

            elif "+" in clauses:
                # suffixed version
                vs = clauses.rstrip("+")
                version = cls.version_class(vs)
                is_stable = is_even(version.value.minor)

                if is_stable:
                    # we have a start and end in stable ranges
                    start_version = cls.version_class(vs)
                    end_version = cls.version_class(str(start_version.value.next_minor()))
                    vstart = VersionConstraint(comparator=">=", version=start_version)
                    vend = VersionConstraint(comparator="<", version=end_version)
                    constraints.extend([vstart, vend])
                else:
                    # mainline branch ranges are resolved to a single constraint
                    version = cls.version_class(vs)
                    constraint = VersionConstraint(comparator=">=", version=version)
                    constraints.append(constraint)

            else:
                # plain single version
                version = cls.version_class(clauses)
                constraint = VersionConstraint(comparator="=", version=version)
                constraints.append(constraint)

        return cls(constraints=constraints)


class OpensslVersionRange(VersionRange):
    """
    Openssl version range.
    openssl doesn't use <,>,<= or >=
    For more see 'https://www.openssl.org/news/vulnerabilities.xml'

    For example::
    >>> from univers.versions import OpensslVersion
    >>> constraints = (
    ... VersionConstraint(version=OpensslVersion("1.0.1af")),
    ... VersionConstraint(comparator="=", version=OpensslVersion("3.0.1")),
    ... VersionConstraint(comparator="=", version=OpensslVersion("1.1.1nf")),
    ... )
    >>> range = OpensslVersionRange(constraints=constraints)
    >>> assert str(range) == 'vers:openssl/1.0.1af|1.1.1nf|3.0.1'
    """

    scheme = "openssl"
    version_class = versions.OpensslVersion

    @classmethod
    def from_native(cls, string):
        cleaned = remove_spaces(string).lower()
        constraints = []
        for version in cleaned.split(","):
            version_obj = cls.version_class(version)
            constraint = VersionConstraint(comparator="=", version=version_obj)
            constraints.append(constraint)
        return cls(constraints=constraints)


vers_by_github_native_comparators = {
    "=": "=",
    "!=": "!=",
    "<=": "<=",
    ">=": ">=",
    "<": "<",
    ">": ">",
}


def build_constraint_from_github_advisory_string(scheme: str, string: str):
    """
    Return a VersionConstraint built from a single github-native version
    relationship ``string``.

    >>> vr = build_constraint_from_github_advisory_string("gem","<= 2.24")
    >>> assert str(vr) == "<=2.24", str(vr)
    >>> vr = build_constraint_from_github_advisory_string("pypi","< 9.0")
    >>> assert str(vr) == "<9.0", str(vr)
    >>> vr = build_constraint_from_github_advisory_string("pypi","= 9.0")
    >>> assert str(vr) == "9.0", str(vr)
    """
    vrc = RANGE_CLASS_BY_SCHEMES[scheme]
    comparator, version = split_req(
        string=string,
        comparators=vers_by_github_native_comparators,
    )
    version = vrc.version_class(version)
    return VersionConstraint(comparator=comparator, version=version)


def build_range_from_github_advisory_constraint(scheme: str, string: str):
    """
    Github has a special syntax for version ranges.
    For example:
    Maven native version range looks like:
    ``[1.0.0,1.0.1)``
    Github native version range looks like:
    ``>= 1.0.0, < 1.0.1``

    Return a VersionRange built from a ``string`` single github-native
    version relationship string.

    For example::

    >>> vr = build_range_from_github_advisory_constraint("maven", ">= 2.13.0, < 2.16.0")
    >>> assert str(vr) == "vers:maven/>=2.13.0|<2.16.0"
    >>> vr = build_range_from_github_advisory_constraint("gem", ">= 2.13.0, < 2.16.0")
    >>> assert str(vr) == "vers:gem/>=2.13.0|<2.16.0"
    >>> vr = build_range_from_github_advisory_constraint("pypi","< 9.0")
    >>> assert str(vr) == "vers:pypi/<9.0"
    >>> vr = build_range_from_github_advisory_constraint("pypi","= 9.0")
    >>> assert str(vr) == "vers:pypi/9.0"
    """
    constraint_strings = string.split(",")
    constraints = []
    vrc = RANGE_CLASS_BY_SCHEMES[scheme]
    for constraint in constraint_strings:
        constraints.append(build_constraint_from_github_advisory_string(scheme, constraint))
    return vrc(constraints=constraints)


def is_even(s):
    """
    Return True if the string "s" is an even number and False if this is an odd
    number. For example:

    >>> is_even(4)
    True
    >>> is_even(123)
    False
    >>> is_even(0)
    True
    """
    return (int(s) % 2) == 0


RANGE_CLASS_BY_SCHEMES = {
    "npm": NpmVersionRange,
    "deb": DebianVersionRange,
    "pypi": PypiVersionRange,
    "maven": MavenVersionRange,
    "nuget": NugetVersionRange,
    "composer": ComposerVersionRange,
    "gem": GemVersionRange,
    "rpm": RpmVersionRange,
    "golang": GolangVersionRange,
    "generic": GenericVersionRange,
    "apache": ApacheVersionRange,
    "hex": HexVersionRange,
    "cargo": CargoVersionRange,
    "mozilla": MozillaVersionRange,
    "github": GitHubVersionRange,
    "ebuild": EbuildVersionRange,
    "archlinux": ArchLinuxVersionRange,
    "nginx": NginxVersionRange,
    "openssl": OpensslVersionRange,
}
